import os
import sys
import unittest
import warnings

import numpy as np

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from sega_learn.time_series.moving_average import (
    ExponentialMovingAverage,
    SimpleMovingAverage,
    WeightedMovingAverage,
)
from tests.utils import BaseTest


class TestSimpleMovingAverage(BaseTest):
    """Unit test suite for the SimpleMovingAverage class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        print("\nTesting SimpleMovingAverage", end="", flush=True)
        warnings.simplefilter("ignore", category=RuntimeWarning)
        warnings.simplefilter("ignore", category=UserWarning)

    def setUp(self):  # NOQA D201
        self.time_series = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], dtype=float)
        self.sma = SimpleMovingAverage(window=3)

    def test_initialization(self):
        """Test SimpleMovingAverage initialization."""
        self.assertEqual(self.sma.window, 3)
        self.assertIsNone(self.sma.smoothed_values)
        self.assertIsNone(self.sma.model)

    def test_initialization_invalid_window(self):
        """Test SimpleMovingAverage initialization with invalid window."""
        with self.assertRaises(ValueError):
            SimpleMovingAverage(window=0)

    def test_fit(self):
        """Test fitting the model."""
        smoothed = self.sma.fit(self.time_series)
        self.assertIsNotNone(smoothed)
        self.assertEqual(len(smoothed), len(self.time_series))
        self.assertTrue(np.isnan(smoothed[:2]).all())  # First 2 values should be NaN

    def test_fit_empty_series(self):
        """Test fitting the model with an empty series."""
        smoothed = self.sma.fit(np.array([]))
        with self.assertRaises((ValueError, TypeError)):
            self.assertEqual(len(smoothed), 0)

    def test_forecast(self):
        """Test forecasting."""
        self.sma.fit(self.time_series)
        forecast = self.sma.forecast(5)
        self.assertEqual(len(forecast), 5)
        self.assertTrue(
            np.all(forecast == forecast[0])
        )  # All values should be the same

    def test_forecast_invalid_steps(self):
        """Test forecasting with invalid steps."""
        self.sma.fit(self.time_series)
        forecast = self.sma.forecast(0)
        self.assertEqual(len(forecast), 0)


class TestWeightedMovingAverage(BaseTest):
    """Unit test suite for the WeightedMovingAverage class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        print("\nTesting WeightedMovingAverage", end="", flush=True)
        warnings.simplefilter("ignore", category=RuntimeWarning)
        warnings.simplefilter("ignore", category=UserWarning)

    def setUp(self):  # NOQA D201
        self.time_series = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], dtype=float)
        self.wma = WeightedMovingAverage(window=3)

    def test_initialization(self):
        """Test WeightedMovingAverage initialization."""
        self.assertEqual(self.wma.window, 3)
        self.assertIsNotNone(self.wma.weights)
        self.assertIsNone(self.wma.smoothed_values)
        self.assertIsNone(self.wma.model)

    def test_initialization_invalid_window(self):
        """Test WeightedMovingAverage initialization with invalid window."""
        with self.assertRaises(ValueError):
            WeightedMovingAverage(window=0)

    def test_initialization_invalid_weights(self):
        """Test WeightedMovingAverage initialization with invalid weights."""
        with self.assertRaises(ValueError):
            WeightedMovingAverage(window=3, weights=[0.5, 0.5])

    def test_fit(self):
        """Test fitting the model."""
        smoothed = self.wma.fit(self.time_series)
        self.assertIsNotNone(smoothed)
        self.assertEqual(len(smoothed), len(self.time_series))
        self.assertTrue(np.isnan(smoothed[:2]).all())  # First 2 values should be NaN

    def test_fit_empty_series(self):
        """Test fitting the model with an empty series."""
        smoothed = self.wma.fit(np.array([]))
        with self.assertRaises((ValueError, TypeError)):
            self.assertEqual(len(smoothed), 0)

    def test_forecast(self):
        """Test forecasting."""
        self.wma.fit(self.time_series)
        forecast = self.wma.forecast(5)
        self.assertEqual(len(forecast), 5)
        self.assertTrue(
            np.all(forecast == forecast[0])
        )  # All values should be the same

    def test_forecast_invalid_steps(self):
        """Test forecasting with invalid steps."""
        self.wma.fit(self.time_series)
        forecast = self.wma.forecast(0)
        self.assertEqual(len(forecast), 0)


class TestExponentialMovingAverage(BaseTest):
    """Unit test suite for the ExponentialMovingAverage class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        print("\nTesting ExponentialMovingAverage", end="", flush=True)
        warnings.simplefilter("ignore", category=RuntimeWarning)
        warnings.simplefilter("ignore", category=UserWarning)

    def setUp(self):  # NOQA D201
        self.time_series = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], dtype=float)
        self.ema = ExponentialMovingAverage(alpha=0.5)

    def test_initialization(self):
        """Test ExponentialMovingAverage initialization."""
        self.assertEqual(self.ema.alpha, 0.5)
        self.assertIsNone(self.ema.smoothed_values)
        self.assertIsNone(self.ema.model)
        self.assertIsNone(self.ema._last_ema)

    def test_initialization_invalid_alpha(self):
        """Test ExponentialMovingAverage initialization with invalid alpha."""
        with self.assertRaises(ValueError):
            ExponentialMovingAverage(alpha=1.5)

    def test_fit(self):
        """Test fitting the model."""
        smoothed = self.ema.fit(self.time_series)
        self.assertIsNotNone(smoothed)
        self.assertEqual(len(smoothed), len(self.time_series))

    def test_fit_empty_series(self):
        """Test fitting the model with an empty series."""
        smoothed = self.ema.fit(np.array([]))
        with self.assertRaises((ValueError, TypeError)):
            self.assertEqual(len(smoothed), 0)

    def test_forecast(self):
        """Test forecasting."""
        self.ema.fit(self.time_series)
        forecast = self.ema.forecast(5)
        self.assertEqual(len(forecast), 5)
        self.assertTrue(
            np.all(forecast == forecast[0])
        )  # All values should be the same

    def test_forecast_invalid_steps(self):
        """Test forecasting with invalid steps."""
        self.ema.fit(self.time_series)
        forecast = self.ema.forecast(0)
        self.assertEqual(len(forecast), 0)


if __name__ == "__main__":
    unittest.main()



import contextlib
import glob
import importlib.util
import io
import os
import sys
import unittest

from tests.utils import BaseTest, strip_file_path

# Change the working directory to the parent directory to allow importing the segadb package.
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))


class TestExampleExceptions(BaseTest):
    """Test cases to check for exceptions in example files."""

    def test_main(self, example_file):
        """Main test function to check for exceptions in example files."""
        pass


class TestExamplesClustering(BaseTest):
    """Test cases for the example files. Holds dynamically generated test cases for each example file."""

    pass


def load_tests(loader, tests, pattern):
    """Dynamically load test cases for each example file.

    Args:
        loader: The test loader instance.
        tests: The test cases to load.
        pattern: The pattern to match test files.
    """
    # Find all example files in the examples directory. (Files starting with 'example_')
    example_files = glob.glob(
        os.path.join(os.path.dirname(__file__), "..\\examples\\clustering\\*.py")
    )

    # Raise an error if no example files are found.
    if not example_files:
        raise FileNotFoundError("No example files found.")

    # Dynamically generate test cases for each example file.
    for example_file in example_files:
        test_name = f"test_{os.path.basename(example_file)}"

        def test_func(self, example_file=example_file):
            """Tests the functionality of a given example file by importing it as a module and executing it."""
            print(
                f"\nTesting example file: {strip_file_path(example_file)}",
                end="",
                flush=True,
            )

            # Import the example file as a module and execute it.
            spec = importlib.util.spec_from_file_location("module.name", example_file)
            example_module = importlib.util.module_from_spec(spec)

            # Redirect stdout to suppress output from the example file.
            with contextlib.redirect_stdout(io.StringIO()):
                spec.loader.exec_module(example_module)

        # Dynamically add the test function to the TestExamples class.
        # If example file runs in __main__, use a lambda function to call the test function.
        test_exeptions = []
        test_exeptions = [f"test_{name}" for name in test_exeptions]
        if test_name in test_exeptions:
            setattr(
                TestExamplesClustering,
                test_name,
                lambda self,
                example_file=example_file: TestExampleExceptions().test_main(
                    example_file
                ),
            )
        else:
            setattr(TestExamplesClustering, test_name, test_func)

    # Load the dynamically generated test cases.
    return loader.loadTestsFromTestCase(TestExamplesClustering)


if __name__ == "__main__":
    unittest.main()



import os
import sys
import unittest
import warnings

import numpy as np

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from sega_learn.time_series.arima import ARIMA, SARIMA, SARIMAX
from tests.utils import BaseTest


# --- Helper function for creating seasonal data ---
def create_seasonal_data(length=100, period=12, amplitude=10, trend=0.1, noise_std=2):
    """Creates sample time series data with seasonality, trend, and noise."""
    time = np.arange(length)
    seasonal_component = amplitude * np.sin(2 * np.pi * time / period)
    trend_component = trend * time
    noise_component = np.random.normal(0, noise_std, length)
    return (
        trend_component + seasonal_component + noise_component + 50
    )  # Add constant offset


class TestARIMA(BaseTest):
    """Unit test suite for the ARIMA class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        print("\nTesting ARIMA", end="", flush=True)
        # Suppress common warnings during tests (e.g., linalg warnings)
        warnings.simplefilter("ignore", category=RuntimeWarning)
        warnings.simplefilter("ignore", category=UserWarning)

    def setUp(self):  # NOQA D201
        self.time_series = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], dtype=float)
        self.order = (1, 1, 1)  # Use simpler order for basic tests
        self.arima = ARIMA(order=self.order)

    def test_initialization(self):
        """Test ARIMA initialization."""
        self.assertEqual(self.arima.order, self.order)
        self.assertEqual(self.arima.p, self.order[0])
        self.assertEqual(self.arima.d, self.order[1])
        self.assertEqual(self.arima.q, self.order[2])
        self.assertIsNone(self.arima.model)
        self.assertIsNone(self.arima.fitted_model)

    def test_invalid_order_negative_value(self):
        """Test ARIMA initialization with a negative value in the order."""
        with self.assertRaisesRegex(ValueError, "p, d, and q must be non-negative"):
            ARIMA(order=(1, -1, 1))

    def test_invalid_order_missing_parameter(self):
        """Test ARIMA initialization with a missing parameter in the order."""
        with self.assertRaisesRegex(
            ValueError, "Order must be a list or tuple of length 3"
        ):
            ARIMA(order=(1, 1))

    def test_invalid_order_invalid_type(self):
        """Test ARIMA initialization with an invalid type for the order."""
        with self.assertRaisesRegex(
            ValueError, "Order must be a list or tuple of length 3"
        ):
            ARIMA(order="invalid")

    def test_invalid_order_extra_parameter(self):
        """Test ARIMA initialization with an extra parameter in the order."""
        with self.assertRaisesRegex(
            ValueError, "Order must be a list or tuple of length 3"
        ):
            ARIMA(order=(1, 1, 1, 1))

    def test_fit(self):
        """Test the fit method."""
        self.arima.fit(self.time_series)
        self.assertIsNotNone(self.arima.fitted_model)
        self.assertTrue("ar_coefficients" in self.arima.fitted_model)
        self.assertTrue("ma_coefficients" in self.arima.fitted_model)
        self.assertEqual(len(self.arima.fitted_model["ar_coefficients"]), self.arima.p)
        self.assertEqual(len(self.arima.fitted_model["ma_coefficients"]), self.arima.q)
        self.assertIsNotNone(self.arima.model)
        np.testing.assert_array_equal(self.arima.model, self.time_series)

    def test_fit_invalid_data_type(self):
        """Test fitting with invalid data type."""
        # Expecting TypeError or potentially ValueError depending on numpy conversion
        with self.assertRaises((TypeError, ValueError)):
            self.arima.fit("invalid_data")

    def test_fit_insufficient_data_for_differencing(self):
        """Test fitting when differencing degree exceeds data length."""
        short_series = np.array([1, 2], dtype=float)
        arima_d2 = ARIMA(order=(1, 2, 1))
        with self.assertRaisesRegex(
            ValueError,
            "Time series length must be greater than the differencing order d.",
        ):
            arima_d2.fit(short_series)

    def test_forecast(self):
        """Test the forecast method."""
        self.arima.fit(self.time_series)
        forecasted_values = self.arima.forecast(steps=3)
        self.assertEqual(len(forecasted_values), 3)
        self.assertTrue(isinstance(forecasted_values, np.ndarray))

    # TODO: Check this test for linear trend
    # def test_forecast_linear_trend(self):
    #     """Test forecasting with a linear trend."""
    #     # A simple check: ARIMA(0,1,0) on linear trend should predict the trend
    #     arima_010 = ARIMA(order=(0,1,0))
    #     arima_010.fit(self.time_series)
    #     forecast_010 = arima_010.forecast(steps=2)
    #     print(f"Forecast: {forecast_010}")
    #     np.testing.assert_array_almost_equal(forecast_010, [11.0, 12.0])

    def test_forecast_without_fit(self):
        """Test forecasting without fitting the model."""
        with self.assertRaisesRegex(
            ValueError, "model must be fitted before forecasting"
        ):
            self.arima.forecast(steps=3)

    def test_differencing_d0(self):
        """Test the _difference_series method."""
        diff0 = self.arima._difference_series(self.time_series, 0)
        np.testing.assert_array_equal(diff0, self.time_series)

    def test_differencing_d1(self):
        """Test the _difference_series method."""
        diff1 = self.arima._difference_series(self.time_series, 1)
        np.testing.assert_array_equal(diff1, np.ones(len(self.time_series) - 1))

    def test_differencing_d2(self):
        """Test the _difference_series method."""
        diff2 = self.arima._difference_series(self.time_series, 2)
        np.testing.assert_array_equal(diff2, np.zeros(len(self.time_series) - 2))

    # --- Static method tests (require statsmodels) ---
    try:
        import statsmodels

        _has_statsmodels = True
    except ImportError:
        _has_statsmodels = False

    @unittest.skipIf(
        not _has_statsmodels, "statsmodels not installed, skipping suggest_order tests"
    )
    def test_suggest_order(self):
        """Test the suggest_order static method."""
        # On linear trend, d=1 or d=2 should be found, p/q likely 0 or 1
        suggested_order = ARIMA.suggest_order(
            self.time_series, max_p=1, max_d=2, max_q=1
        )
        self.assertEqual(len(suggested_order), 3)
        self.assertTrue(all(isinstance(i, int) for i in suggested_order))
        self.assertIn(suggested_order[1], [1, 2])  # d should be 1 or 2

    @unittest.skipIf(
        not _has_statsmodels, "statsmodels not installed, skipping suggest_order tests"
    )
    def test_suggest_order_invalid_data(self):
        """Test suggest_order with invalid data."""
        # numpy array conversion might handle some cases, but non-numeric should fail
        with self.assertRaises((TypeError, ValueError)):
            ARIMA.suggest_order(["a", "b", "c"])

    @unittest.skipIf(
        not _has_statsmodels, "statsmodels not installed, skipping suggest_order tests"
    )
    def test_suggest_order_empty_data(self):
        """Test suggest_order with empty or too short data."""
        with self.assertRaises(
            ValueError
        ):  # ADF test usually raises on very short series
            ARIMA.suggest_order([])
        with self.assertRaises(ValueError):
            ARIMA.suggest_order([1])

    def test_find_best_order(self):
        """Test the find_best_order static method."""
        train_series = self.time_series[:7]
        test_series = self.time_series[7:]
        # Expect low order like (0,1,0) or similar for linear trend
        best_order = ARIMA.find_best_order(
            train_series, test_series, max_p=1, max_d=2, max_q=1
        )
        self.assertEqual(len(best_order), 3)
        self.assertTrue(all(isinstance(i, int) for i in best_order))
        self.assertTrue(0 <= best_order[0] <= 1)
        self.assertTrue(0 <= best_order[1] <= 2)
        self.assertTrue(0 <= best_order[2] <= 1)
        # Check if (0,1,0) is found as it's optimal here
        # self.assertEqual(best_order, (0, 1, 0)) # May vary slightly based on implementation details

    def test_find_best_order_invalid_data_type(self):
        """Test find_best_order with invalid data type."""
        with self.assertRaisesRegex(ValueError, "must be list or numpy array"):
            ARIMA.find_best_order("invalid", self.time_series[7:])
        with self.assertRaisesRegex(ValueError, "must be list or numpy array"):
            ARIMA.find_best_order(self.time_series[:7], "invalid")

    def test_find_best_order_empty_data(self):
        """Test find_best_order with empty data."""
        with self.assertRaisesRegex(ValueError, "must not be empty"):
            ARIMA.find_best_order([], self.time_series[7:])
        with self.assertRaisesRegex(ValueError, "must not be empty"):
            ARIMA.find_best_order(self.time_series[:7], [])

    def test_find_best_order_subset(self):
        """Test find_best_order with a subset of the training series."""
        long_series = np.linspace(0, 50, 51)
        train_series = long_series[:40]
        test_series = long_series[40:]
        # Ensure subsetting doesn't crash
        best_order = ARIMA.find_best_order(
            train_series, test_series, subset_size=0.5, max_p=1, max_d=1, max_q=1
        )
        self.assertEqual(len(best_order), 3)
        self.assertTrue(all(isinstance(i, int) for i in best_order))

    def test_find_best_order_invalid_subset(self):
        """Test find_best_order with an invalid subset size."""
        train_series = self.time_series[:7]
        test_series = self.time_series[7:]
        with self.assertRaisesRegex(ValueError, "subset_size must be between 0 and 1"):
            ARIMA.find_best_order(train_series, test_series, subset_size=1.5)
        with self.assertRaisesRegex(ValueError, "subset_size must be between 0 and 1"):
            ARIMA.find_best_order(train_series, test_series, subset_size=-0.5)
        with self.assertRaisesRegex(ValueError, "subset_size must be between 0 and 1"):
            ARIMA.find_best_order(train_series, test_series, subset_size=0)


class TestSARIMA(BaseTest):
    """Unit test suite for the SARIMA class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        print("\nTesting SARIMA", end="", flush=True)
        warnings.simplefilter("ignore", category=RuntimeWarning)
        warnings.simplefilter("ignore", category=UserWarning)

    def setUp(self):  # NOQA D201
        self.seasonal_period = 4  # Simple seasonality
        self.time_series = create_seasonal_data(
            length=40, period=self.seasonal_period, trend=0.05, noise_std=1
        )
        self.order = (1, 0, 1)
        self.seasonal_order = (1, 1, 1, self.seasonal_period)  # P, D, Q, m
        self.sarima = SARIMA(order=self.order, seasonal_order=self.seasonal_order)

    def test_initialization(self):
        """Test SARIMA initialization."""
        self.assertEqual(self.sarima.order, self.order)
        self.assertEqual(self.sarima.seasonal_order, self.seasonal_order)
        self.assertEqual(self.sarima.p, self.order[0])
        self.assertEqual(self.sarima.d, self.order[1])
        self.assertEqual(self.sarima.q, self.order[2])
        self.assertEqual(self.sarima.P, self.seasonal_order[0])
        self.assertEqual(self.sarima.D, self.seasonal_order[1])
        self.assertEqual(self.sarima.Q, self.seasonal_order[2])
        self.assertEqual(self.sarima.m, self.seasonal_order[3])
        self.assertIsNone(self.sarima.model)  # Inherited, check parent init
        self.assertIsNone(self.sarima.fitted_model)
        self.assertIsNone(self.sarima.original_series)

    def test_invalid_seasonal_order_length_less_than_4(self):
        """Test SARIMA initialization with invalid seasonal order length."""
        with self.assertRaisesRegex(
            ValueError, "Seasonal order must be a tuple/list of length 4"
        ):
            SARIMA(seasonal_order=(1, 1, 1))

    def test_invalid_seasonal_order_length_more_than_4(self):
        """Test SARIMA initialization with invalid seasonal order length."""
        with self.assertRaisesRegex(
            ValueError, "Seasonal order must be a tuple/list of length 4"
        ):
            SARIMA(seasonal_order=(1, 1, 1, 4, 5))

    def test_invalid_seasonal_order_values(self):
        """Test SARIMA initialization with invalid seasonal order values."""
        with self.assertRaisesRegex(
            ValueError, "P, D, Q must be ≥0 and m must be a positive integer."
        ):
            SARIMA(seasonal_order=(-1, 1, 1, 4))
        with self.assertRaisesRegex(
            ValueError, "P, D, Q must be ≥0 and m must be a positive integer."
        ):
            SARIMA(seasonal_order=(1, -1, 1, 4))
        with self.assertRaisesRegex(
            ValueError, "P, D, Q must be ≥0 and m must be a positive integer."
        ):
            SARIMA(seasonal_order=(1, 1, -1, 4))
        with self.assertRaisesRegex(
            ValueError, "P, D, Q must be ≥0 and m must be a positive integer."
        ):
            SARIMA(seasonal_order=(1, 1, 1, 0))
        with self.assertRaisesRegex(
            ValueError, "P, D, Q must be ≥0 and m must be a positive integer."
        ):
            SARIMA(seasonal_order=(1, 1, 1, -4))

    def test_seasonal_difference_d1(self):
        """Test the _seasonal_difference method."""
        m = self.seasonal_period
        D1_diff = self.sarima._seasonal_difference(self.time_series, D=1, m=m)
        self.assertEqual(len(D1_diff), len(self.time_series) - m)
        expected_D1 = self.time_series[m:] - self.time_series[:-m]
        np.testing.assert_array_almost_equal(D1_diff, expected_D1)

    def test_seasonal_difference_d2(self):
        """Test the _seasonal_difference method."""
        m = self.seasonal_period
        D1_diff = self.sarima._seasonal_difference(self.time_series, D=1, m=m)
        self.assertEqual(len(D1_diff), len(self.time_series) - m)
        expected_D1 = self.time_series[m:] - self.time_series[:-m]
        # D=2
        D2_diff = self.sarima._seasonal_difference(self.time_series, D=2, m=m)
        self.assertEqual(len(D2_diff), len(self.time_series) - 2 * m)
        expected_D2 = expected_D1[m:] - expected_D1[:-m]
        np.testing.assert_array_almost_equal(D2_diff, expected_D2)

    def test_seasonal_difference_d0(self):
        """Test the _seasonal_difference method."""
        m = self.seasonal_period
        # D=0
        D0_diff = self.sarima._seasonal_difference(self.time_series, D=0, m=m)
        np.testing.assert_array_almost_equal(D0_diff, self.time_series)

    # TODO: Fix this test
    # def test_inverse_seasonal_difference_roundtrip(self):
    #     """Test _inverse_seasonal_difference by round-tripping."""
    #     # This is hard to test directly in isolation as it needs history.
    #     # We test it implicitly via fit/forecast.
    #     # Let's try a simple round trip though
    #     sarima_000_0104 = SARIMA(order=(0,0,0), seasonal_order=(0,1,0,4))
    #     sarima_000_0104.original_series = self.time_series # Manually set for test
    #     sarima_000_0104.D = 1
    #     sarima_000_0104.m = 4
    #     diff_series = sarima_000_0104._seasonal_difference(self.time_series, D=1, m=4)

    #     # The inverse function forecasts step-by-step, so simulate that
    #     # We need to predict the *differenced* values first
    #     # Let's just use the known differenced values as 'forecasts'
    #     reconstructed = sarima_000_0104._inverse_seasonal_difference(diff_series)

    #     # Should reconstruct the original series starting from index m
    #     np.testing.assert_array_almost_equal(reconstructed, self.time_series[sarima_000_0104.m:])

    def test_fit(self):
        """Test the SARIMA fit method."""
        self.sarima.fit(self.time_series)
        self.assertIsNotNone(self.sarima.fitted_model)  # Check ARIMA part got fitted
        self.assertTrue("ar_coefficients" in self.sarima.fitted_model)
        self.assertTrue("ma_coefficients" in self.sarima.fitted_model)
        self.assertIsNotNone(self.sarima.original_series)
        np.testing.assert_array_equal(self.sarima.original_series, self.time_series)
        # Check if differencing was applied (model length should be shorter if D or d > 0)
        # Length after seasonal diff D=1: N - m
        # Length after non-seasonal diff d=0: N - m
        # Length fed to _fit_ar_model: N - m
        # Length of y in _fit_ar_model: N - m - p
        # Check internal model length implicitly via forecast
        self.assertIsNotNone(
            self.sarima.model
        )  # ARIMA's model (the potentially differenced series)

    # TODO: Fix this test
    # def test_fit_no_seasonal_differencing(self):
    #     """Test fit when D=0."""
    #     sarima_no_sd = SARIMA(order=(1,1,0), seasonal_order=(1,0,0,4))
    #     sarima_no_sd.fit(self.time_series)
    #     self.assertIsNotNone(sarima_no_sd.fitted_model)
    #     self.assertEqual(sarima_no_sd.D, 0)
    #     # The model fitted should be on the non-seasonally differenced series
    #     expected_model_input = sarima_no_sd._difference_series(self.time_series, d=1)
    #     np.testing.assert_array_almost_equal(sarima_no_sd.model, expected_model_input)

    def test_forecast(self):
        """Test the SARIMA forecast method."""
        self.sarima.fit(self.time_series)
        steps = 5
        forecasted_values = self.sarima.forecast(steps=steps)
        self.assertEqual(len(forecasted_values), steps)
        self.assertTrue(isinstance(forecasted_values, np.ndarray))
        # Check values are reasonable (not NaN or Inf)
        self.assertTrue(np.all(np.isfinite(forecasted_values)))

    def test_forecast_without_fit(self):
        """Test forecasting without fitting the model."""
        with self.assertRaisesRegex(ValueError, "Fit the model before forecasting"):
            self.sarima.forecast(steps=3)

    # --- Static method tests (require statsmodels) ---
    @unittest.skipIf(
        not TestARIMA._has_statsmodels,
        "statsmodels not installed, skipping SARIMA suggest/find tests",
    )
    def test_suggest_order(self):
        """Test the SARIMA suggest_order static method."""
        suggested_orders = SARIMA.suggest_order(
            self.time_series, max_m=self.seasonal_period + 1
        )  # Limit m for speed
        self.assertEqual(len(suggested_orders), 2)
        self.assertEqual(len(suggested_orders[0]), 3)  # (p, d, q)
        self.assertEqual(len(suggested_orders[1]), 4)  # (P, D, Q, m)
        self.assertTrue(all(isinstance(i, int) for i in suggested_orders[0]))
        self.assertTrue(all(isinstance(i, int) for i in suggested_orders[1]))
        # Check if detected 'm' is reasonable (should be close to self.seasonal_period or 1)
        detected_m = suggested_orders[1][3]
        self.assertTrue(detected_m > 0)
        # self.assertEqual(detected_m, self.seasonal_period) # Might not always detect perfectly

    @unittest.skipIf(
        not TestARIMA._has_statsmodels,
        "statsmodels not installed, skipping SARIMA suggest/find tests",
    )
    def test_find_best_order(self):
        """Test the SARIMA find_best_order static method."""
        train_series = self.time_series[:-5]
        test_series = self.time_series[-5:]
        best_orders = SARIMA.find_best_order(
            train_series,
            test_series,
            max_p=1,
            max_d=1,
            max_q=1,
            max_P=1,
            max_D=1,
            max_Q=1,
            max_m=self.seasonal_period,  # Fix m for test speed
        )
        self.assertEqual(len(best_orders), 2)
        self.assertEqual(len(best_orders[0]), 3)  # (p, d, q)
        self.assertEqual(len(best_orders[1]), 4)  # (P, D, Q, m)
        self.assertTrue(all(isinstance(i, int) for i in best_orders[0]))
        self.assertTrue(all(isinstance(i, int) for i in best_orders[1]))


class TestSARIMAX(BaseTest):
    """Unit test suite for the SARIMAX class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        print("\nTesting SARIMAX", end="", flush=True)
        warnings.simplefilter("ignore", category=RuntimeWarning)
        warnings.simplefilter("ignore", category=UserWarning)

    def setUp(self):  # NOQA D201
        self.seasonal_period = 4
        self.n_samples = 40
        self.endog = create_seasonal_data(
            length=self.n_samples, period=self.seasonal_period, trend=0.05, noise_std=1
        )
        # Create exogenous variable strongly correlated with endog
        self.exog = self.endog * 0.5 + np.random.normal(0, 5, self.n_samples)
        self.exog = self.exog.reshape(-1, 1)  # Ensure 2D

        self.order = (1, 0, 0)
        self.seasonal_order = (1, 1, 0, self.seasonal_period)
        self.sarimax = SARIMAX(order=self.order, seasonal_order=self.seasonal_order)

        # For forecasting
        self.steps = 5
        self.exog_future = np.random.rand(self.steps, 1) * 10 + 50  # Future exog values

    def test_initialization(self):
        """Test SARIMAX initialization."""
        self.assertEqual(self.sarimax.order, self.order)
        self.assertEqual(self.sarimax.seasonal_order, self.seasonal_order)
        self.assertEqual(self.sarimax.p, self.order[0])
        self.assertEqual(self.sarimax.P, self.seasonal_order[0])
        self.assertEqual(self.sarimax.m, self.seasonal_order[3])
        self.assertIsNone(self.sarimax.beta)
        self.assertIsNone(self.sarimax.k_exog)
        self.assertIsNone(self.sarimax.fitted_model)  # From parent
        self.assertIsNone(self.sarimax.original_series)  # From parent

    def test_fit(self):
        """Test the SARIMAX fit method."""
        self.sarimax.fit(self.endog, self.exog)

        # Check beta and k_exog
        self.assertIsNotNone(self.sarimax.beta)
        self.assertIsNotNone(self.sarimax.k_exog)
        self.assertEqual(self.sarimax.k_exog, self.exog.shape[1])
        self.assertEqual(len(self.sarimax.beta), self.exog.shape[1])

        # Check underlying SARIMA fit
        self.assertIsNotNone(self.sarimax.fitted_model)
        self.assertTrue("ar_coefficients" in self.sarimax.fitted_model)
        self.assertTrue("ma_coefficients" in self.sarimax.fitted_model)
        self.assertIsNotNone(
            self.sarimax.original_series
        )  # Should store original endog
        np.testing.assert_array_equal(self.sarimax.original_series, self.endog)

    def test_fit_1d_exog(self):
        """Test fitting with 1D exogenous data."""
        exog_1d = self.exog.flatten()
        self.sarimax.fit(self.endog, exog_1d)
        self.assertEqual(self.sarimax.k_exog, 1)
        self.assertEqual(len(self.sarimax.beta), 1)
        self.assertIsNotNone(self.sarimax.fitted_model)

    def test_fit_mismatched_lengths(self):
        """Test fit with endog and exog having different lengths."""
        with self.assertRaisesRegex(ValueError, "y and exog must have the same length"):
            self.sarimax.fit(self.endog, self.exog[:-1])

    def test_forecast(self):
        """Test the SARIMAX forecast method."""
        self.sarimax.fit(self.endog, self.exog)
        forecasted_values = self.sarimax.forecast(
            steps=self.steps, exog_future=self.exog_future
        )

        self.assertEqual(len(forecasted_values), self.steps)
        self.assertTrue(isinstance(forecasted_values, np.ndarray))
        self.assertTrue(np.all(np.isfinite(forecasted_values)))

    def test_forecast_1d_exog_future(self):
        """Test forecasting with 1D future exogenous data."""
        self.sarimax.fit(self.endog, self.exog)
        # Requires k_exog = 1
        sarimax_k1 = SARIMAX(order=(0, 0, 0), seasonal_order=(0, 0, 0, 1))
        exog_k1 = self.exog[:, 0]  # Use first column as 1D
        sarimax_k1.fit(self.endog, exog_k1)
        self.assertEqual(sarimax_k1.k_exog, 1)

        exog_future_1d = self.exog_future.flatten()
        forecast = sarimax_k1.forecast(steps=self.steps, exog_future=exog_future_1d)
        self.assertEqual(len(forecast), self.steps)

    def test_forecast_without_fit(self):
        """Test SARIMAX forecasting without fitting."""
        with self.assertRaisesRegex(ValueError, "Fit model before forecasting"):
            self.sarimax.forecast(steps=self.steps, exog_future=self.exog_future)

    def test_forecast_mismatched_exog_future_shape(self):
        """Test forecast with incorrect number of columns in exog_future."""
        self.sarimax.fit(self.endog, self.exog)
        # Fit with k_exog = 1
        self.assertEqual(self.sarimax.k_exog, 1)
        # Try forecasting with 2 columns
        exog_future_bad = np.random.rand(self.steps, 2)
        with self.assertRaisesRegex(
            ValueError, f"exog_future must have {self.sarimax.k_exog} columns"
        ):
            self.sarimax.forecast(steps=self.steps, exog_future=exog_future_bad)

        # Fit with k_exog = 2
        exog2 = np.hstack([self.exog, self.exog * 2])
        sarimax2 = SARIMAX(self.order, self.seasonal_order)
        sarimax2.fit(self.endog, exog2)
        self.assertEqual(sarimax2.k_exog, 2)
        # Try forecasting with 1 column
        exog_future_bad2 = np.random.rand(self.steps, 1)
        with self.assertRaisesRegex(
            ValueError, f"exog_future must have {sarimax2.k_exog} columns"
        ):
            sarimax2.forecast(steps=self.steps, exog_future=exog_future_bad2)

    # --- Static method tests (require statsmodels) ---
    @unittest.skipIf(
        not TestARIMA._has_statsmodels,
        "statsmodels not installed, skipping SARIMAX suggest/find tests",
    )
    def test_suggest_order(self):
        """Test the SARIMAX suggest_order static method."""
        suggested_orders = SARIMAX.suggest_order(
            self.endog, self.exog, max_m=self.seasonal_period + 1
        )
        self.assertEqual(len(suggested_orders), 2)
        self.assertEqual(len(suggested_orders[0]), 3)  # (p, d, q)
        self.assertEqual(len(suggested_orders[1]), 4)  # (P, D, Q, m)
        self.assertTrue(all(isinstance(i, int) for i in suggested_orders[0]))
        self.assertTrue(all(isinstance(i, int) for i in suggested_orders[1]))

    @unittest.skipIf(
        not TestARIMA._has_statsmodels,
        "statsmodels not installed, skipping SARIMAX suggest/find tests",
    )
    def test_find_best_order(self):
        """Test the SARIMAX find_best_order static method."""
        split_idx = self.n_samples - self.steps
        train_endog, test_endog = self.endog[:split_idx], self.endog[split_idx:]
        train_exog, test_exog = self.exog[:split_idx], self.exog[split_idx:]

        best_orders = SARIMAX.find_best_order(
            train_endog,
            test_endog,
            train_exog,
            test_exog,
            max_p=1,
            max_d=0,
            max_q=0,
            max_P=1,
            max_D=1,
            max_Q=0,
            max_m=self.seasonal_period,  # Limit search space
        )
        self.assertEqual(len(best_orders), 2)
        self.assertEqual(len(best_orders[0]), 3)  # (p, d, q)
        self.assertEqual(len(best_orders[1]), 4)  # (P, D, Q, m)
        self.assertTrue(all(isinstance(i, int) for i in best_orders[0]))
        self.assertTrue(all(isinstance(i, int) for i in best_orders[1]))


if __name__ == "__main__":
    unittest.main()



import os
import sys
import unittest

import numpy as np

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from sega_learn.neural_networks import *
from tests.utils import BaseTest


class MockLayer:
    """Mock layer class for testing optimizers."""

    def __init__(self, input_size, output_size):
        """Initialize the MockLayer class."""
        self.weights = np.random.randn(input_size, output_size) * 0.01
        self.biases = np.zeros((1, output_size))


class TestAdamOptimizer(BaseTest):
    """Unit tests for the AdamOptimizer class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        """Print message before running tests."""
        print("\nTesting AdamOptimizer", end="", flush=True)

    def setUp(self):  # NOQA D201
        """Initialize the AdamOptimizer class and layers."""
        self.optimizer = AdamOptimizer()
        self.layers = [MockLayer(3, 2), MockLayer(2, 1)]
        self.optimizer.initialize(self.layers)

    def test_initialize(self):
        """Test the initialize method."""
        self.assertEqual(len(self.optimizer.m), len(self.layers))
        self.assertEqual(len(self.optimizer.v), len(self.layers))
        for m, v, layer in zip(
            self.optimizer.m, self.optimizer.v, self.layers, strict=False
        ):
            np.testing.assert_array_equal(m, np.zeros_like(layer.weights))
            np.testing.assert_array_equal(v, np.zeros_like(layer.weights))

    def test_update(self):
        """Test the update method."""
        layer = self.layers[0]
        dW = np.random.randn(*layer.weights.shape)
        db = np.random.randn(*layer.biases.shape)
        initial_weights = layer.weights.copy()
        initial_biases = layer.biases.copy()
        self.optimizer.update(layer, dW, db, 0)
        self.assertFalse(np.array_equal(layer.weights, initial_weights))
        self.assertFalse(np.array_equal(layer.biases, initial_biases))

    def test_update_with_zero_gradients(self):
        """Test update method with zero gradients."""
        layer = self.layers[0]
        dW = np.zeros_like(layer.weights)
        db = np.zeros_like(layer.biases)
        initial_weights = layer.weights.copy()
        initial_biases = layer.biases.copy()
        self.optimizer.update(layer, dW, db, 0)
        np.testing.assert_array_equal(layer.weights, initial_weights)
        np.testing.assert_array_equal(layer.biases, initial_biases)

    def test_update_with_large_gradients(self):
        """Test update method with large gradients."""
        layer = self.layers[0]
        dW = np.ones_like(layer.weights) * 1e5
        db = np.ones_like(layer.biases) * 1e5
        self.optimizer.update(layer, dW, db, 0)
        self.assertTrue(np.all(layer.weights < 1e5))
        self.assertTrue(np.all(layer.biases < 1e5))

    def test_update_with_small_gradients(self):
        """Test update method with small gradients."""
        layer = self.layers[0]
        dW = np.ones_like(layer.weights) * 1e-5
        db = np.ones_like(layer.biases) * 1e-5
        initial_weights = layer.weights.copy()
        initial_biases = layer.biases.copy()
        self.optimizer.update(layer, dW, db, 0)
        self.assertFalse(np.array_equal(layer.weights, initial_weights))
        self.assertFalse(np.array_equal(layer.biases, initial_biases))


class TestSGDOptimizer(BaseTest):
    """Unit tests for the SGDOptimizer class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        """Print message before running tests."""
        print("\nTesting SGDOptimizer", end="", flush=True)

    def setUp(self):  # NOQA D201
        """Initialize the SGDOptimizer class and layers."""
        self.optimizer = SGDOptimizer()
        self.layers = [MockLayer(3, 2), MockLayer(2, 1)]
        self.optimizer.initialize(self.layers)

    def test_initialize(self):
        """Test the initialize method."""
        self.assertEqual(len(self.optimizer.velocity), len(self.layers))
        for v, layer in zip(self.optimizer.velocity, self.layers, strict=False):
            np.testing.assert_array_equal(v, np.zeros_like(layer.weights))

    def test_update(self):
        """Test the update method."""
        layer = self.layers[0]
        dW = np.random.randn(*layer.weights.shape)
        db = np.random.randn(*layer.biases.shape)
        initial_weights = layer.weights.copy()
        initial_biases = layer.biases.copy()
        self.optimizer.update(layer, dW, db, 0)
        self.assertFalse(np.array_equal(layer.weights, initial_weights))
        self.assertFalse(np.array_equal(layer.biases, initial_biases))

    def test_update_with_zero_gradients(self):
        """Test update method with zero gradients."""
        layer = self.layers[0]
        dW = np.zeros_like(layer.weights)
        db = np.zeros_like(layer.biases)
        initial_weights = layer.weights.copy()
        initial_biases = layer.biases.copy()
        self.optimizer.update(layer, dW, db, 0)
        np.testing.assert_array_equal(layer.weights, initial_weights)
        np.testing.assert_array_equal(layer.biases, initial_biases)

    def test_update_with_large_gradients(self):
        """Test update method with large gradients."""
        layer = self.layers[0]
        dW = np.ones_like(layer.weights) * 1e5
        db = np.ones_like(layer.biases) * 1e5
        self.optimizer.update(layer, dW, db, 0)
        self.assertTrue(np.all(layer.weights < 1e5))
        self.assertTrue(np.all(layer.biases < 1e5))

    def test_update_with_small_gradients(self):
        """Test update method with small gradients."""
        layer = self.layers[0]
        dW = np.ones_like(layer.weights) * 1e-5
        db = np.ones_like(layer.biases) * 1e-5
        initial_weights = layer.weights.copy()
        initial_biases = layer.biases.copy()
        self.optimizer.update(layer, dW, db, 0)
        self.assertFalse(np.array_equal(layer.weights, initial_weights))
        self.assertFalse(np.array_equal(layer.biases, initial_biases))


class TestAdadeltaOptimizer(BaseTest):
    """Unit tests for the AdadeltaOptimizer class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        """Print message before running tests."""
        print("\nTesting AdadeltaOptimizer", end="", flush=True)

    def setUp(self):  # NOQA D201
        """Initialize the AdadeltaOptimizer class and layers."""
        self.optimizer = AdadeltaOptimizer()
        self.layers = [MockLayer(3, 2), MockLayer(2, 1)]
        self.optimizer.initialize(self.layers)

    def test_initialize(self):
        """Test the initialize method."""
        self.assertEqual(len(self.optimizer.E_g2), len(self.layers))
        self.assertEqual(len(self.optimizer.E_delta_x2), len(self.layers))
        for E_g2, E_delta_x2, layer in zip(
            self.optimizer.E_g2, self.optimizer.E_delta_x2, self.layers, strict=False
        ):
            np.testing.assert_array_equal(E_g2, np.zeros_like(layer.weights))
            np.testing.assert_array_equal(E_delta_x2, np.zeros_like(layer.weights))

    def test_update(self):
        """Test the update method."""
        layer = self.layers[0]
        dW = np.random.randn(*layer.weights.shape)
        db = np.random.randn(*layer.biases.shape)
        initial_weights = layer.weights.copy()
        initial_biases = layer.biases.copy()
        self.optimizer.update(layer, dW, db, 0)
        self.assertFalse(np.array_equal(layer.weights, initial_weights))
        self.assertFalse(np.array_equal(layer.biases, initial_biases))

    def test_update_with_zero_gradients(self):
        """Test update method with zero gradients."""
        layer = self.layers[0]
        dW = np.zeros_like(layer.weights)
        db = np.zeros_like(layer.biases)
        initial_weights = layer.weights.copy()
        initial_biases = layer.biases.copy()
        self.optimizer.update(layer, dW, db, 0)
        np.testing.assert_array_equal(layer.weights, initial_weights)
        np.testing.assert_array_equal(layer.biases, initial_biases)

    def test_update_with_large_gradients(self):
        """Test update method with large gradients."""
        layer = self.layers[0]
        dW = np.ones_like(layer.weights) * 1e5
        db = np.ones_like(layer.biases) * 1e5
        self.optimizer.update(layer, dW, db, 0)
        self.assertTrue(np.all(layer.weights < 1e5))
        self.assertTrue(np.all(layer.biases < 1e5))

    def test_update_with_small_gradients(self):
        """Test update method with small gradients."""
        layer = self.layers[0]
        dW = np.ones_like(layer.weights) * 1e-5
        db = np.ones_like(layer.biases) * 1e-5
        initial_weights = layer.weights.copy()
        initial_biases = layer.biases.copy()
        self.optimizer.update(layer, dW, db, 0)
        self.assertFalse(np.array_equal(layer.weights, initial_weights))
        self.assertFalse(np.array_equal(layer.biases, initial_biases))


if __name__ == "__main__":
    unittest.main()



import cupy as cp
from cupy import fuse


# Fused dropout kernel: combines masking and scaling into one operation.
@fuse()
def fused_dropout(x, dropout_rate, random_vals):
    return cp.where(random_vals < (1 - dropout_rate), x / (1 - dropout_rate), 0)


def apply_dropout(X, dropout_rate):
    # Generate dropout mask and apply fused dropout.
    random_vals = cp.random.rand(*X.shape)
    return fused_dropout(X, dropout_rate, random_vals)


# Fused activation functions to fuse operations and reduce kernel launches.
@fuse()
def fused_relu(x):
    return cp.where(x > 0, x, 0)


@fuse()
def fused_sigmoid(x):
    return 1 / (1 + cp.exp(-x))


@fuse()
def fused_leaky_relu(x, alpha=0.01):
    return cp.where(x > 0, x, alpha * x)


# Optimized forward pass with in-place operations and fused kernels.
def forward_cupy(X, weights, biases, activations, dropout_rate, training, is_binary):
    num_layers = len(weights)
    layer_outputs = [X]
    for i in range(num_layers - 1):
        # Compute linear transformation.
        Z = cp.dot(layer_outputs[i], weights[i]) + biases[i]

        # Apply activation using in-place or fused operations.
        if activations[i] == "relu":
            cp.maximum(Z, 0, out=Z)  # In-place ReLU.
            A = Z
        elif activations[i] == "leaky_relu":
            A = fused_leaky_relu(Z)
        elif activations[i] == "tanh":
            cp.tanh(Z, out=Z)  # In-place tanh.
            A = Z
        elif activations[i] == "sigmoid":
            A = fused_sigmoid(Z)
        elif activations[i] == "softmax":
            # Use log-sum-exp trick for numerical stability.
            max_Z = cp.max(Z, axis=1, keepdims=True)
            Z -= max_Z  # In-place subtraction.
            cp.exp(Z, out=Z)
            sum_Z = cp.sum(Z, axis=1, keepdims=True)
            A = Z / sum_Z
        else:
            raise ValueError(f"Unsupported activation: {activations[i]}")

        # Apply dropout if in training mode.
        if training and dropout_rate > 0:
            A = apply_dropout(A, dropout_rate)
        layer_outputs.append(A)

    # Output layer computation.
    Z = cp.dot(layer_outputs[-1], weights[-1]) + biases[-1]
    if is_binary:
        A = fused_sigmoid(Z)
    else:
        max_Z = cp.max(Z, axis=1, keepdims=True)
        Z -= max_Z
        cp.exp(Z, out=Z)
        sum_Z = cp.sum(Z, axis=1, keepdims=True)
        A = Z / sum_Z
    layer_outputs.append(A)
    return layer_outputs


# Backward pass remains similar; note that you can also consider fusing derivative computations.
def backward_cupy(
    layer_outputs, y, weights, activations, reg_lambda, is_binary, dWs, dbs
):
    m = y.shape[0]
    num_layers = len(weights)
    outputs = layer_outputs[-1]
    if is_binary:
        y = y.reshape(-1, 1).astype(cp.float64)
        dA = -(y / (outputs + 1e-15) - (1 - y) / (1 - outputs + 1e-15))
    else:
        dA = outputs.copy()
        dA[cp.arange(m), y] -= 1  # Advanced indexing supported in CuPy.

    for i in range(num_layers - 1, -1, -1):
        prev_activation = layer_outputs[i]
        if i < num_layers - 1:
            output = layer_outputs[i + 1]
            if activations[i] == "relu":
                dZ = dA * (output > 0)
            elif activations[i] == "leaky_relu":
                dZ = dA * cp.where(output > 0, 1, 0.01)
            elif activations[i] == "tanh":
                dZ = dA * (1 - output**2)
            elif activations[i] == "sigmoid":
                dZ = dA * output * (1 - output)
            elif activations[i] == "softmax":
                dZ = dA
            else:
                raise ValueError(f"Unsupported activation: {activations[i]}")
        else:
            dZ = dA

        dW = cp.dot(prev_activation.T, dZ) / m + reg_lambda * weights[i]
        db = cp.sum(dZ, axis=0, keepdims=True) / m
        dWs[i] = dW
        dbs[i] = db
        if i > 0:
            dA = cp.dot(dZ, weights[i].T)
    return dWs, dbs


# Optimized loss functions.
def logsumexp(a, axis=None, keepdims=False):
    a_max = cp.max(a, axis=axis, keepdims=True)
    out = cp.log(cp.sum(cp.exp(a - a_max), axis=axis, keepdims=True)) + a_max
    if not keepdims:
        out = cp.squeeze(out, axis=axis)
    return out


def calculate_cross_entropy_loss(logits, targets):
    n = logits.shape[0]
    c_i = cp.argmax(targets, axis=1)
    loss = cp.mean(logsumexp(logits, axis=1) - logits[cp.arange(n), c_i])
    return loss


def calculate_bce_with_logits_loss(logits, targets):
    # Numerically stable BCE with logits formulation.
    loss = cp.mean(
        cp.maximum(logits, 0) - logits * targets + cp.log1p(cp.exp(-cp.abs(logits)))
    )
    return loss


def calculate_loss_from_outputs_binary(outputs, y, weights, reg_lambda):
    loss = calculate_bce_with_logits_loss(outputs, y)
    if isinstance(weights, list):
        l2_reg = reg_lambda * cp.sum([cp.sum(w**2) for w in weights])
        loss += l2_reg
    return float(loss)


def calculate_loss_from_outputs_multi(outputs, y, weights, reg_lambda):
    loss = calculate_cross_entropy_loss(outputs, y)
    if isinstance(weights, list):
        l2_reg = reg_lambda * cp.sum([cp.sum(w**2) for w in weights])
        loss += l2_reg
    return float(loss)


def evaluate_batch(y_hat, y_true, is_binary):
    if is_binary:
        predicted = (y_hat > 0.5).astype(cp.int32).ravel()
        accuracy = cp.mean(predicted == y_true.ravel())
    else:
        predicted = cp.argmax(y_hat, axis=1).astype(cp.int32)
        accuracy = cp.mean(predicted == y_true)
    return accuracy



import os
import sys
import unittest
import warnings

import numpy as np
from numpy.testing import assert_allclose, assert_array_equal

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from sega_learn.time_series.decomposition import (
    AdditiveDecomposition,
    MultiplicativeDecomposition,
    _centered_moving_average,
)
from tests.utils import BaseTest


# --- Helper function for creating seasonal data ---
def create_seasonal_data(
    length=48, period=12, amplitude=10, trend=0.1, noise_std=2, offset=50
):
    """Creates sample time series data with seasonality, trend, and noise."""
    time = np.arange(length)
    seasonal_component = amplitude * np.sin(2 * np.pi * time / period)
    trend_component = trend * time
    noise_component = np.random.normal(0, noise_std, length)
    return trend_component + seasonal_component + noise_component + offset


class TestAdditiveDecomposition(BaseTest):
    """Unit test suite for the AdditiveDecomposition class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        print("\nTesting AdditiveDecomposition", end="", flush=True)
        warnings.simplefilter("ignore", category=RuntimeWarning)
        warnings.simplefilter("ignore", category=UserWarning)

    def setUp(self):  # NOQA D201
        self.period = 12
        self.time_series = create_seasonal_data(
            length=4 * self.period, period=self.period
        )
        self.decomp = AdditiveDecomposition(period=self.period)

    def test_initialization(self):
        """Test AdditiveDecomposition initialization."""
        self.assertEqual(self.decomp.period, self.period)
        self.assertIsNone(self.decomp.time_series)
        self.assertIsNone(self.decomp.trend)
        self.assertIsNone(self.decomp.seasonal)
        self.assertIsNone(self.decomp.residual)

    def test_initialization_invalid_period_low(self):
        """Test initialization with invalid period (<= 1)."""
        with self.assertRaises(ValueError):
            AdditiveDecomposition(period=1)
        with self.assertRaises(ValueError):
            AdditiveDecomposition(period=0)

    def test_initialization_invalid_period_type(self):
        """Test initialization with non-integer period."""
        with self.assertRaises(ValueError):
            AdditiveDecomposition(period=12.5)
        with self.assertRaises(ValueError):
            AdditiveDecomposition(period="12")

    def test_fit(self):
        """Test fitting the model."""
        self.decomp.fit(self.time_series)
        n = len(self.time_series)
        self.assertIsNotNone(self.decomp.time_series)
        self.assertIsNotNone(self.decomp.trend)
        self.assertIsNotNone(self.decomp.seasonal)
        self.assertIsNotNone(self.decomp.residual)
        self.assertEqual(len(self.decomp.trend), n)
        self.assertEqual(len(self.decomp.seasonal), n)
        self.assertEqual(len(self.decomp.residual), n)
        self.assertIsInstance(self.decomp.trend, np.ndarray)
        self.assertIsInstance(self.decomp.seasonal, np.ndarray)
        self.assertIsInstance(self.decomp.residual, np.ndarray)
        # Check that seasonal component sums approximately to zero over one period
        self.assertTrue(
            np.isclose(np.sum(self.decomp.seasonal[: self.period]), 0, atol=1e-8)
        )

    def test_fit_short_series_even_period(self):
        """Test fitting with series too short for even period CMA."""
        short_series = self.time_series[: self.period]
        with self.assertRaises(ValueError):
            self.decomp.fit(short_series)

    def test_fit_short_series_odd_period(self):
        """Test fitting with series too short for odd period CMA."""
        period = 7
        decomp_odd = AdditiveDecomposition(period=period)
        short_series = self.time_series[
            : period - 1
        ]  # Length 6, needs 7 for period 7 CMA
        with self.assertRaises(ValueError):
            decomp_odd.fit(short_series)

    def test_fit_nan_series(self):
        """Test fitting with NaNs in the series."""
        nan_series = self.time_series.copy()
        nan_series[5] = np.nan
        nan_series[20] = np.nan
        with self.assertWarns(UserWarning):
            self.decomp.fit(nan_series)
        n = len(nan_series)
        self.assertEqual(len(self.decomp.trend), n)
        self.assertEqual(len(self.decomp.seasonal), n)
        self.assertEqual(len(self.decomp.residual), n)
        # Expect NaNs in components corresponding to input NaNs and CMA edge effects
        self.assertTrue(np.isnan(self.decomp.trend).any())
        self.assertTrue(np.isnan(self.decomp.residual).any())
        # Seasonal might be complete if enough data exists per season despite NaNs
        # self.assertTrue(np.isnan(self.decomp.seasonal).any()) # This depends on NaN positions

    def test_get_components(self):
        """Test retrieving components after fitting."""
        self.decomp.fit(self.time_series)
        components = self.decomp.get_components()
        self.assertIsInstance(components, dict)
        self.assertIn("trend", components)
        self.assertIn("seasonal", components)
        self.assertIn("residual", components)
        self.assertIs(components["trend"], self.decomp.trend)

    def test_get_components_before_fit(self):
        """Test calling get_components before fitting."""
        with self.assertRaisesRegex(ValueError, "Model has not been fitted yet."):
            self.decomp.get_components()

    def test_reconstruct(self):
        """Test reconstructing the series from components."""
        self.decomp.fit(self.time_series)
        reconstructed = self.decomp.reconstruct()
        self.assertEqual(len(reconstructed), len(self.time_series))
        # Check if reconstruction is close to original where trend is not NaN
        valid_indices = ~np.isnan(self.decomp.trend)
        self.assertTrue(
            np.allclose(
                reconstructed[valid_indices], self.time_series[valid_indices], atol=1e-8
            )
        )


class TestMultiplicativeDecomposition(BaseTest):
    """Unit test suite for the MultiplicativeDecomposition class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        print("\nTesting MultiplicativeDecomposition", end="", flush=True)
        warnings.simplefilter("ignore", category=RuntimeWarning)
        warnings.simplefilter("ignore", category=UserWarning)

    def setUp(self):  # NOQA D201
        self.period = 12
        # Data strictly positive
        self.time_series_pos = create_seasonal_data(
            length=4 * self.period, period=self.period, offset=50
        )
        # Data with non-positive values
        self.time_series_nonpos = create_seasonal_data(
            length=4 * self.period, period=self.period, offset=-10
        )

        self.decomp_pos = MultiplicativeDecomposition(period=self.period)
        self.decomp_nonpos = MultiplicativeDecomposition(period=self.period)

    def test_initialization(self):
        """Test MultiplicativeDecomposition initialization."""
        self.assertEqual(self.decomp_pos.period, self.period)
        self.assertIsNone(self.decomp_pos.time_series)
        self.assertEqual(self.decomp_pos.offset, 0.0)
        self.assertIsNone(self.decomp_pos.trend)
        self.assertIsNone(self.decomp_pos.seasonal)
        self.assertIsNone(self.decomp_pos.residual)
        self.assertIsNone(self.decomp_pos._trend_shifted)

    def test_initialization_invalid_period_low(self):
        """Test initialization with invalid period (<= 1)."""
        with self.assertRaises(ValueError):
            MultiplicativeDecomposition(period=1)

    def test_initialization_invalid_period_type(self):
        """Test initialization with non-integer period."""
        with self.assertRaises(ValueError):
            MultiplicativeDecomposition(period=7.5)

    # --- Tests using strictly positive data ---

    def test_fit_positive(self):
        """Test fitting the model with positive data."""
        self.decomp_pos.fit(self.time_series_pos)
        n = len(self.time_series_pos)
        self.assertIsNotNone(self.decomp_pos.time_series)
        self.assertIsNotNone(self.decomp_pos.trend)
        self.assertIsNotNone(self.decomp_pos.seasonal)
        self.assertIsNotNone(self.decomp_pos.residual)
        self.assertEqual(self.decomp_pos.offset, 0.0)  # No offset applied
        self.assertEqual(len(self.decomp_pos.trend), n)
        self.assertEqual(len(self.decomp_pos.seasonal), n)
        self.assertEqual(len(self.decomp_pos.residual), n)
        # Check that seasonal component averages approximately to 1 over one period
        self.assertTrue(
            np.isclose(np.mean(self.decomp_pos.seasonal[: self.period]), 1.0, atol=1e-8)
        )

    def test_fit_positive_short_series(self):
        """Test fitting positive data with series too short."""
        short_series = self.time_series_pos[: self.period]
        self.decomp_pos.fit(short_series)

    def test_fit_positive_nan_series(self):
        """Test fitting positive data with NaNs."""
        nan_series = self.time_series_pos.copy()
        nan_series[5] = np.nan
        self.decomp_pos.fit(nan_series)
        n = len(nan_series)
        self.assertEqual(len(self.decomp_pos.trend), n)
        self.assertEqual(len(self.decomp_pos.seasonal), n)
        self.assertEqual(len(self.decomp_pos.residual), n)
        self.assertEqual(self.decomp_pos.offset, 0.0)
        self.assertTrue(np.isnan(self.decomp_pos.trend).any())
        self.assertTrue(np.isnan(self.decomp_pos.residual).any())

    def test_get_components_positive(self):
        """Test retrieving components after fitting positive data."""
        self.decomp_pos.fit(self.time_series_pos)
        components = self.decomp_pos.get_components()
        self.assertIsInstance(components, dict)
        self.assertIn("trend", components)
        self.assertIn("seasonal", components)
        self.assertIn("residual", components)

    def test_get_components_before_fit(self):
        """Test calling get_components before fitting."""
        with self.assertRaisesRegex(ValueError, "Model has not been fitted yet."):
            self.decomp_pos.get_components()

    def test_reconstruct_positive(self):
        """Test reconstructing the series from components (positive fit)."""
        self.decomp_pos.fit(self.time_series_pos)
        reconstructed = self.decomp_pos.reconstruct()
        self.assertEqual(len(reconstructed), len(self.time_series_pos))
        valid_indices = (
            ~np.isnan(self.decomp_pos.trend)
            & ~np.isnan(self.decomp_pos.seasonal)
            & ~np.isnan(self.decomp_pos.residual)
        )
        # Reconstruction might not be perfect due to approximations in decomposition
        self.assertTrue(
            np.allclose(
                reconstructed[valid_indices],
                self.time_series_pos[valid_indices],
                rtol=0.1,
            )
        )  # Use relative tolerance

    # --- Tests using non-positive data ---

    def test_fit_non_positive(self):
        """Test fitting the model with non-positive data."""
        with self.assertWarns(UserWarning):
            self.decomp_nonpos.fit(self.time_series_nonpos)
            # Check if the offset warning was issued

        n = len(self.time_series_nonpos)
        self.assertIsNotNone(self.decomp_nonpos.time_series)
        self.assertIsNotNone(self.decomp_nonpos.trend)
        self.assertIsNotNone(self.decomp_nonpos.seasonal)
        self.assertIsNotNone(self.decomp_nonpos.residual)
        self.assertGreater(self.decomp_nonpos.offset, 0.0)  # Offset must be applied
        self.assertEqual(len(self.decomp_nonpos.trend), n)
        self.assertEqual(len(self.decomp_nonpos.seasonal), n)
        self.assertEqual(len(self.decomp_nonpos.residual), n)
        # Seasonal component should still average approximately to 1
        self.assertTrue(
            np.isclose(
                np.mean(self.decomp_nonpos.seasonal[: self.period]), 1.0, atol=1e-8
            )
        )
        # Trend component might be negative now
        self.assertTrue(
            np.any(self.decomp_nonpos.trend[~np.isnan(self.decomp_nonpos.trend)] < 0)
        )

    def test_fit_non_positive_short_series(self):
        """Test fitting non-positive data with series too short."""
        short_series = self.time_series_nonpos[: self.period]
        self.decomp_nonpos.fit(short_series)

    def test_get_components_non_positive(self):
        """Test retrieving components after fitting non-positive data."""
        self.decomp_nonpos.fit(self.time_series_nonpos)
        components = self.decomp_nonpos.get_components()
        self.assertIsInstance(components, dict)
        self.assertIn("trend", components)  # Trend is adjusted back
        self.assertIn("seasonal", components)  # Seasonal is from shifted
        self.assertIn("residual", components)  # Residual is from shifted

    def test_reconstruct_non_positive(self):
        """Test reconstructing the series from components (non-positive fit)."""
        self.decomp_nonpos.fit(self.time_series_nonpos)
        reconstructed = self.decomp_nonpos.reconstruct()
        self.assertEqual(len(reconstructed), len(self.time_series_nonpos))
        valid_indices = (
            ~np.isnan(self.decomp_nonpos.trend)
            & ~np.isnan(self.decomp_nonpos.seasonal)
            & ~np.isnan(self.decomp_nonpos.residual)
        )
        # Check closeness to the *original* non-positive series
        # Due to the offset and multiplicative nature, might not be perfectly close
        self.assertTrue(
            np.allclose(
                reconstructed[valid_indices],
                self.time_series_nonpos[valid_indices],
                rtol=0.2,
                atol=1.0,
            )
        )  # Looser tolerance


class TestCenteredMovingAverage(BaseTest):
    """Unit test suite for the _centered_moving_average helper function."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        print("\nTesting _centered_moving_average", end="", flush=True)
        warnings.simplefilter("ignore", category=RuntimeWarning)

    # --- Test Cases ---

    def test_odd_window(self):
        """Test CMA with an odd window size."""
        series = np.array([1, 2, 3, 4, 5, 6, 7], dtype=float)
        window = 3
        expected = np.array([np.nan, 2.0, 3.0, 4.0, 5.0, 6.0, np.nan])
        result = _centered_moving_average(series, window)
        assert_allclose(result, expected, equal_nan=True, rtol=1e-8, atol=1e-8)

    def test_odd_window_longer(self):
        """Test CMA with an odd window size on a longer series."""
        series = np.arange(1, 11, dtype=float)  # [1, 2, ..., 10]
        window = 5
        expected = np.array(
            [np.nan, np.nan, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, np.nan, np.nan]
        )
        result = _centered_moving_average(series, window)
        assert_allclose(result, expected, equal_nan=True, rtol=1e-8, atol=1e-8)

    def test_even_window(self):
        """Test CMA with an even window size."""
        series = np.array([1, 2, 3, 4, 5, 6, 7, 8], dtype=float)
        window = 4
        expected = np.array([np.nan, np.nan, 3.0, 4.0, 5.0, 6.0, np.nan, np.nan])
        result = _centered_moving_average(series, window)
        assert_allclose(result, expected, equal_nan=True, rtol=1e-8, atol=1e-8)

    def test_even_window_longer(self):
        """Test CMA with an even window size on a longer series."""
        series = np.arange(1, 13, dtype=float)  # [1, 2, ..., 12]
        window = 6
        expected = np.array(
            [
                np.nan,
                np.nan,
                np.nan,
                4.0,
                5.0,
                6.0,
                7.0,
                8.0,
                9.0,
                np.nan,
                np.nan,
                np.nan,
            ]
        )
        result = _centered_moving_average(series, window)
        assert_allclose(result, expected, equal_nan=True, rtol=1e-8, atol=1e-8)

    def test_window_larger_than_series(self):
        """Test CMA when window size exceeds series length."""
        series = np.array([1, 2, 3], dtype=float)
        window = 5
        expected = np.array([np.nan, np.nan, np.nan])
        # Function should now handle this before convolve
        result = _centered_moving_average(series, window)
        assert_allclose(result, expected, equal_nan=True)  # Corrected expected

    def test_window_equals_series_odd(self):
        """Test CMA when odd window size equals series length."""
        series = np.array([1, 2, 3, 4, 5], dtype=float)
        window = 5
        expected = np.array([np.nan, np.nan, 3.0, np.nan, np.nan])
        result = _centered_moving_average(series, window)
        assert_allclose(result, expected, equal_nan=True)

    def test_window_equals_series_even(self):
        """Test CMA when even window size equals series length."""
        series = np.array([1, 2, 3, 4], dtype=float)
        window = 4
        expected = np.array([np.nan, np.nan, np.nan, np.nan])
        # Function should now handle this case correctly
        result = _centered_moving_average(series, window)
        assert_allclose(result, expected, equal_nan=True)  # Corrected expected

    def test_window_one(self):
        """Test CMA with window size 1."""
        series = np.array([1, 5, 3, 7], dtype=float)
        window = 1
        expected = series.copy()
        result = _centered_moving_average(series, window)
        assert_allclose(result, expected, equal_nan=True)

    def test_window_two(self):
        """Test CMA with window size 2."""
        series = np.array([1, 2, 3, 4, 5], dtype=float)
        window = 2
        expected = np.array([np.nan, 2.0, 3.0, 4.0, np.nan])
        result = _centered_moving_average(series, window)
        assert_allclose(result, expected, equal_nan=True, rtol=1e-8, atol=1e-8)

    def test_series_with_nans(self):
        """Test CMA with NaNs present in the input series."""
        series = np.array([1, 2, np.nan, 4, 5, 6, np.nan, 8, 9], dtype=float)
        window = 3
        # np.convolve propagates NaNs within its calculation window
        # Window 1 (1,2,nan) -> nan (centered at index 1)
        # Window 2 (2,nan,4) -> nan (centered at index 2)
        # Window 3 (nan,4,5) -> nan (centered at index 3)
        # Window 4 (4,5,6) -> 5.0 (centered at index 4)
        # Window 5 (5,6,nan) -> nan (centered at index 5)
        # Window 6 (6,nan,8) -> nan (centered at index 6)
        # Window 7 (nan,8,9) -> nan (centered at index 7)
        expected = np.array(
            [np.nan, np.nan, np.nan, np.nan, 5.0, np.nan, np.nan, np.nan, np.nan]
        )
        result = _centered_moving_average(series, window)
        assert_allclose(result, expected, equal_nan=True)  # Corrected expected

    def test_empty_series(self):
        """Test CMA with an empty input series."""
        series = np.array([], dtype=float)
        window = 3
        expected = np.array([], dtype=float)
        # Function should now handle this before convolve
        result = _centered_moving_average(series, window)
        assert_array_equal(result, expected)

    def test_invalid_window_zero(self):
        """Test CMA with window size 0."""
        series = np.array([1, 2, 3], dtype=float)
        with self.assertRaisesRegex(ValueError, "window must be a positive integer"):
            _centered_moving_average(series, 0)

    def test_invalid_window_negative(self):
        """Test CMA with negative window size."""
        series = np.array([1, 2, 3], dtype=float)
        with self.assertRaisesRegex(ValueError, "window must be a positive integer"):
            _centered_moving_average(series, -3)


if __name__ == "__main__":
    unittest.main()



import os
import sys
import unittest

# Change the working directory to the parent directory to allow importing the package.
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from sega_learn.svm import *
from tests.utils import BaseTest


class TestImportsSVM(BaseTest):
    """Tests that the SVM subpackage can be imported correctly.

    Methods:
    - setUpClass: Initializes a new instance of the Index class before each test method is run.
    - test_individual_imports: Tests that each module in the segadb package can be imported individually.
    - test_wildcard_import: Tests that the segadb package can be imported using a wildcard import.
    """

    @classmethod
    def setUpClass(cls):  # NOQA D201
        print("\nTesting Imports - SVM", end="", flush=True)

    def test_individual_imports(self):
        """Tests that each module in the segadb package can be imported individually."""
        from sega_learn.svm import BaseSVM as bsvm
        from sega_learn.svm import GeneralizedSVC as gsvc
        from sega_learn.svm import GeneralizedSVR as gsvr
        from sega_learn.svm import LinearSVC as lsvc
        from sega_learn.svm import LinearSVR as lsvr
        from sega_learn.svm import OneClassSVM as osvm

        assert bsvm is not None
        assert lsvc is not None
        assert lsvr is not None
        assert osvm is not None
        assert gsvr is not None
        assert gsvc is not None

    def test_wildcard_import(self):
        """Tests that the segadb package can be imported using a wildcard import."""
        assert BaseSVM is not None
        assert LinearSVC is not None
        assert LinearSVR is not None
        assert OneClassSVM is not None
        assert GeneralizedSVR is not None
        assert GeneralizedSVC is not None

    def test_LinearSVC(self):  # NOQA D201
        from sega_learn.svm import LinearSVC

        assert LinearSVC is not None

    def test_LinearSVR(self):  # NOQA D201
        from sega_learn.svm import LinearSVR

        assert LinearSVR is not None

    def test_OneClassSVM(self):  # NOQA D201
        from sega_learn.svm import OneClassSVM

        assert OneClassSVM is not None

    def test_GeneralizedSVC(self):  # NOQA D201
        from sega_learn.svm import GeneralizedSVC

        assert GeneralizedSVC is not None

    def test_GeneralizedSVR(self):  # NOQA D201
        from sega_learn.svm import GeneralizedSVR

        assert GeneralizedSVR is not None

    def test_BaseSVM(self):  # NOQA D201
        from sega_learn.svm import BaseSVM

        assert BaseSVM is not None


if __name__ == "__main__":
    unittest.main()



import os
import sys
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
import numpy as np
from sega_learn.linear_models import *
from sega_learn.linear_models import make_sample_data
from sega_learn.utils import Metrics, make_classification
from tests.utils import BaseTest, suppress_print, synthetic_data_regression

r2_score = Metrics.r_squared
accuracy_score = Metrics.accuracy


class TestOrdinaryLeastSquares(BaseTest):
    """Unit test for the Ordinary Least Squares regression class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        """Initializes a new instance of the Index class before each test method is run."""
        print("\nTesting Ordinary Least Squares Model", end="", flush=True)

    def setUp(self):  # NOQA D201
        self.model = OrdinaryLeastSquares()

    def test_fit_predict(self):
        """Tests the fit and predict methods of the Ordinary Least Squares class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model.fit(X, y)
        y_pred = self.model.predict(X)
        self.assertGreater(r2_score(y, y_pred), 0.5)

    def test_fit_intercept(self):
        """Tests the fit_intercept parameter of the Ordinary Least Squares class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = OrdinaryLeastSquares(fit_intercept=True)
        self.model.fit(X, y)
        self.assertTrue(self.model.fit_intercept)

    def test_invalid_fit(self):
        """Tests the fit method with invalid input."""
        with self.assertRaises(ValueError):
            self.model.fit(None, None)

    def test_invalid_predict(self):
        """Tests the predict method with invalid input."""
        with self.assertRaises(AttributeError):
            self.model.predict(None)

    def test_get_formula(self):
        """Tests the get_formula method of the Ordinary Least Squares class."""
        X, y = synthetic_data_regression(n_samples=100, n_features=2, noise=0.1)
        self.model.fit(X, y)
        formula = self.model.get_formula()
        self.assertIsNotNone(formula)
        self.assertIn("y = ", formula)

    def test_get_formula_with_intercept(self):
        """Tests the get_formula method with intercept of the Ordinary Least Squares class."""
        X, y = synthetic_data_regression(n_samples=100, n_features=2, noise=0.1)
        self.model.fit(X, y)
        formula = self.model.get_formula()
        self.assertIsNotNone(formula)
        self.assertIn("y = ", formula)

    def test_coef_(self):
        """Tests the coef_ attribute of the Ordinary Least Squares class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model.fit(X, y)
        self.assertEqual(len(self.model.coef_), 5)


class TestRidge(BaseTest):
    """Unit test for the Ridge regression class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        """Initializes a new instance of the Index class before each test method is run."""
        print("\nTesting Ridge Regression Model", end="", flush=True)

    def setUp(self):  # NOQA D201
        self.model = Ridge()

    def test_fit_predict(self):
        """Tests the fit and predict methods of the Ridge class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model.fit(X, y)
        y_pred = self.model.predict(X)
        self.assertGreater(r2_score(y, y_pred), 0.5)

    def test_fit_intercept(self):
        """Tests the fit_intercept parameter of the Ridge class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = Ridge(fit_intercept=True)
        self.model.fit(X, y)
        self.assertTrue(self.model.fit_intercept)

    def test_fit_max_iter(self):
        """Tests the max_iter parameter of the Ridge class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = Ridge(max_iter=1000)
        self.model.fit(X, y)
        self.assertEqual(self.model.max_iter, 1000)

    def test_fit_tol(self):
        """Tests the tol parameter of the Ridge class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = Ridge(tol=0.0001)
        self.model.fit(X, y)
        self.assertEqual(self.model.tol, 0.0001)

    def test_invalid_fit(self):
        """Tests the fit method with invalid input."""
        with self.assertRaises(ValueError):
            self.model.fit(None, None)

    def test_invalid_predict(self):
        """Tests the predict method with invalid input."""
        with self.assertRaises(AttributeError):
            self.model.predict(None)

    def test_get_formula(self):
        """Tests the get_formula method of the Ridge class."""
        X, y = synthetic_data_regression(n_samples=100, n_features=2, noise=0.1)
        self.model.fit(X, y)
        formula = self.model.get_formula()
        self.assertIsNotNone(formula)
        self.assertIn("y = ", formula)

    def test_get_formula_with_intercept(self):
        """Tests the get_formula method with intercept of the Ridge class."""
        X, y = synthetic_data_regression(n_samples=100, n_features=2, noise=0.1)
        self.model.fit(X, y)
        formula = self.model.get_formula()
        self.assertIsNotNone(formula)
        self.assertIn("y = ", formula)

    def test_coef_(self):
        """Tests the coef_ attribute of the Ridge class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model.fit(X, y)
        self.assertEqual(len(self.model.coef_), 5)

    def test_fit_predict_numba(self):
        """Test the fit and predict methods with numba implementation."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = Ridge(alpha=1.0, fit_intercept=True, max_iter=1000, tol=1e-4)
        self.model.fit(X, y, numba=True)
        y_pred = self.model.predict(X)
        self.assertGreater(r2_score(y, y_pred), 0.5)

    def test_fit_no_intercept_numba(self):
        """Test the fit method with numba implementation and no intercept."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = Ridge(alpha=1.0, fit_intercept=False, max_iter=1000, tol=1e-4)
        self.model.fit(X, y, numba=True)
        self.assertEqual(self.model.intercept_, 0.0)

    def test_invalid_fit_numba(self):
        """Test the fit method with invalid input using numba implementation."""
        self.model = Ridge(alpha=1.0, fit_intercept=True, max_iter=1000, tol=1e-4)
        with self.assertRaises(ValueError):
            self.model.fit(None, None, numba=True)


class TestLasso(BaseTest):
    """Unit test for the Lasso regression class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        """Initializes a new instance of the Index class before each test method is run."""
        print("\nTesting Lasso Regression Model", end="", flush=True)

    def setUp(self):  # NOQA D201
        self.model = Lasso()

    def test_fit_predict(self):
        """Tests the fit and predict methods of the Lasso class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model.fit(X, y)
        y_pred = self.model.predict(X)
        self.assertGreater(r2_score(y, y_pred), 0.5)

    def test_fit_intercept(self):
        """Tests the fit_intercept parameter of the Lasso class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = Lasso(fit_intercept=True)
        self.model.fit(X, y)
        self.assertTrue(self.model.fit_intercept)

    def test_fit_max_iter(self):
        """Tests the max_iter parameter of the Lasso class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = Lasso(max_iter=1000)
        self.model.fit(X, y)
        self.assertEqual(self.model.max_iter, 1000)

    def test_fit_tol(self):
        """Tests the tol parameter of the Lasso class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = Lasso(tol=0.0001)
        self.model.fit(X, y)
        self.assertEqual(self.model.tol, 0.0001)

    def test_invalid_fit(self):
        """Tests the fit method with invalid input."""
        with self.assertRaises(ValueError):
            self.model.fit(None, None)

    def test_invalid_predict(self):
        """Tests the predict method with invalid input."""
        with self.assertRaises(AttributeError):
            self.model.predict(None)

    def test_get_formula(self):
        """Tests the get_formula method of the Lasso class."""
        X, y = synthetic_data_regression(n_samples=100, n_features=2, noise=0.1)
        self.model.fit(X, y)
        formula = self.model.get_formula()
        self.assertIsNotNone(formula)
        self.assertIn("y = ", formula)

    def test_get_formula_with_intercept(self):
        """Tests the get_formula method with intercept of the Lasso class."""
        X, y = synthetic_data_regression(n_samples=100, n_features=2, noise=0.1)
        self.model.fit(X, y)
        formula = self.model.get_formula()
        self.assertIsNotNone(formula)
        self.assertIn("y = ", formula)

    def test_coef_(self):
        """Tests the coef_ attribute of the Lasso class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model.fit(X, y)
        self.assertEqual(len(self.model.coef_), 5)

    def test_fit_predict_numba(self):
        """Test the fit and predict methods with numba implementation."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = Lasso(alpha=1.0, fit_intercept=True, max_iter=1000, tol=1e-4)
        self.model.fit(X, y, numba=True)
        y_pred = self.model.predict(X)
        self.assertGreater(r2_score(y, y_pred), 0.5)

    def test_fit_no_intercept_numba(self):
        """Test the fit method with numba implementation and no intercept."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = Lasso(alpha=1.0, fit_intercept=False, max_iter=1000, tol=1e-4)
        self.model.fit(X, y, numba=True)
        self.assertEqual(self.model.intercept_, 0.0)

    def test_invalid_fit_numba(self):
        """Test the fit method with invalid input using numba implementation."""
        self.model = Lasso(alpha=1.0, fit_intercept=True, max_iter=1000, tol=1e-4)
        with self.assertRaises(ValueError):
            self.model.fit(None, None, numba=True)


class TestBayesian(BaseTest):
    """Unit test for the Bayesian regression class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        """Initializes a new instance of the Index class before each test method is run."""
        print("\nTesting Bayesian Regression Model", end="", flush=True)

    def setUp(self):  # NOQA D201
        self.model = Bayesian()

    def test_fit_predict(self):
        """Tests the fit and predict methods of the Bayesian class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model.fit(X, y)
        y_pred = self.model.predict(X)
        self.assertGreater(r2_score(y, y_pred), 0.5)

    def test_fit_intercept(self):
        """Tests the fit_intercept parameter of the Bayesian class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = Bayesian(fit_intercept=True)
        self.model.fit(X, y)
        self.assertTrue(self.model.fit_intercept)

    def test_fit_max_iter(self):
        """Tests the max_iter parameter of the Bayesian class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = Bayesian(max_iter=1000)
        self.model.fit(X, y)
        self.assertEqual(self.model.max_iter, 1000)

    def test_fit_tol(self):
        """Tests the tol parameter of the Bayesian class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = Bayesian(tol=0.0001)
        self.model.fit(X, y)
        self.assertEqual(self.model.tol, 0.0001)

    def test_tune(self):
        """Tests the tune method of the Bayesian class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        with suppress_print():
            alpha_1, alpha_2, lambda_1, lambda_2 = self.model.tune(
                X, y, beta1=0.9, beta2=0.999, iter=1000
            )
        self.assertIsNotNone(alpha_1)
        self.assertIsNotNone(alpha_2)
        self.assertIsNotNone(lambda_1)
        self.assertIsNotNone(lambda_2)

    def test_invalid_fit(self):
        """Tests the fit method with invalid input."""
        with self.assertRaises(ValueError):
            self.model.fit(None, None)

    def test_invalid_predict(self):
        """Tests the predict method with invalid input."""
        with self.assertRaises(TypeError):
            self.model.predict(None)

    def test_get_formula(self):
        """Tests the get_formula method of the Bayesian class."""
        X, y = synthetic_data_regression(n_samples=100, n_features=2, noise=0.1)
        self.model.fit(X, y)
        formula = self.model.get_formula()
        self.assertIsNotNone(formula)
        self.assertIn("y = ", formula)

    def test_get_formula_with_intercept(self):
        """Tests the get_formula method with intercept of the Bayesian class."""
        X, y = synthetic_data_regression(n_samples=100, n_features=2, noise=0.1)
        self.model.fit(X, y)
        formula = self.model.get_formula()
        self.assertIsNotNone(formula)
        self.assertIn("y = ", formula)

    def test_coef_(self):
        """Tests the coef_ attribute of the Bayesian class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model.fit(X, y)
        self.assertEqual(len(self.model.coef_), 5)


class TestRANSAC(BaseTest):
    """Unit test for the RANSAC regression class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        """Initializes a new instance of the Index class before each test method is run."""
        print("\nTesting RANSAC Regression Model", end="", flush=True)

    def setUp(self):  # NOQA D201
        self.model = RANSAC()

    def test_fit_predict(self):
        """Tests the fit and predict methods of the RANSAC class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model.fit(X, y)
        y_pred = self.model.predict(X)
        self.assertGreater(r2_score(y, y_pred), 0.5)

    def test_fit_n(self):
        """Tests the n parameter of the RANSAC class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = RANSAC(n=100)
        self.model.fit(X, y)
        self.assertEqual(self.model.n, 100)

    def test_fit_k(self):
        """Tests the k parameter of the RANSAC class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = RANSAC(k=10)
        self.model.fit(X, y)
        self.assertEqual(self.model.k, 10)

    def test_fit_t(self):
        """Tests the t parameter of the RANSAC class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = RANSAC(t=2.0)
        self.model.fit(X, y)
        self.assertEqual(self.model.t, 2.0)

    def test_fit_d(self):
        """Tests the d parameter of the RANSAC class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = RANSAC(d=10)
        self.model.fit(X, y)
        self.assertEqual(self.model.d, 10)

    def test_fit_auto_scale_t(self):
        """Tests the auto_scale_t parameter of the RANSAC class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = RANSAC(auto_scale_t=True)
        self.model.fit(X, y)
        self.assertTrue(self.model.scale_threshold)

    def test_fit_auto_scale_n(self):
        """Tests the auto_scale_n parameter of the RANSAC class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = RANSAC(auto_scale_n=True)
        self.model.fit(X, y)
        self.assertTrue(self.model.scale_n)

    def test_invalid_fit(self):
        """Tests the fit method with invalid input."""
        with self.assertRaises(ValueError):
            self.model.fit(None, None)

    def test_invalid_predict(self):
        """Tests the predict method with invalid input."""
        with self.assertRaises(AttributeError):
            self.model.predict(None)

    def test_get_formula(self):
        """Tests the get_formula method of the RANSAC class."""
        X, y = synthetic_data_regression(n_samples=100, n_features=2, noise=0.1)
        self.model.fit(X, y)
        formula = self.model.get_formula()
        self.assertIsNotNone(formula)
        self.assertIn("y = ", formula)

    def test_get_formula_with_intercept(self):
        """Tests the get_formula method with intercept of the RANSAC class."""
        X, y = synthetic_data_regression(n_samples=100, n_features=2, noise=0.1)
        self.model.fit(X, y)
        formula = self.model.get_formula()
        self.assertIsNotNone(formula)
        self.assertIn("y = ", formula)


class TestPassiveAggressiveRegressor(BaseTest):
    """Unit test for the Passive Aggressive Regressor class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        """Initializes a new instance of the Index class before each test method is run."""
        print("\nTesting Passive Aggressive Regressor Model", end="", flush=True)

    def setUp(self):  # NOQA D201
        self.model = PassiveAggressiveRegressor()

    def test_fit_predict(self):
        """Tests the fit and predict methods of the Passive Aggressive Regressor class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        with suppress_print():
            self.model.fit(X, y)
        y_pred = self.model.predict(X)
        self.assertGreater(r2_score(y, y_pred), 0.5)

    def test_fit_all_steps(self):
        """Tests the fit method with save_steps parameter of the Passive Aggressive Regressor class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        with suppress_print():
            self.model.fit(X, y, save_steps=True)
        self.assertTrue(len(self.model.steps_) > 0)

    def test_predict_all_steps(self):
        """Tests the predict_all_steps method of the Passive Aggressive Regressor class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        with suppress_print():
            self.model.fit(X, y, save_steps=True)
        y_preds = self.model.predict_all_steps(X)
        self.assertGreater(len(y_preds), 0)

    def test_fit_max_iter(self):
        """Tests the max_iter parameter of the Passive Aggressive Regressor class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = PassiveAggressiveRegressor(max_iter=1000)
        with suppress_print():
            self.model.fit(X, y)
        self.assertEqual(self.model.max_iter, 1000)

    def test_fit_tol(self):
        """Tests the tol parameter of the Passive Aggressive Regressor class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        self.model = PassiveAggressiveRegressor(tol=0.0001)
        with suppress_print():
            self.model.fit(X, y)
        self.assertEqual(self.model.tol, 0.0001)

    def test_invalid_fit(self):
        """Tests the fit method with invalid input."""
        with self.assertRaises(ValueError):
            self.model.fit(None, None)

    def test_invalid_predict(self):
        """Tests the predict method with invalid input."""
        with self.assertRaises(TypeError):
            self.model.predict(None)

    def test_get_formula(self):
        """Tests the get_formula method of the Passive Aggressive Regressor class."""
        X, y = synthetic_data_regression(n_samples=100, n_features=2, noise=0.1)
        with suppress_print():
            self.model.fit(X, y)
        formula = self.model.get_formula()
        self.assertIsNotNone(formula)
        self.assertIn("y = ", formula)

    def test_get_formula_with_intercept(self):
        """Tests the get_formula method with intercept of the Passive Aggressive Regressor class."""
        X, y = synthetic_data_regression(n_samples=100, n_features=2, noise=0.1)
        with suppress_print():
            self.model.fit(X, y)
        formula = self.model.get_formula()
        self.assertIsNotNone(formula)
        self.assertIn("y = ", formula)

    def test_coef_(self):
        """Tests the coef_ attribute of the Passive Aggressive Regressor class."""
        X, y = synthetic_data_regression(n_samples=1000, n_features=5, noise=0.1)
        with suppress_print():
            self.model.fit(X, y)
        self.assertEqual(len(self.model.coef_), 5)


class TestLinearDiscriminantAnalysis(BaseTest):
    """Unit test for the Linear Discriminant Analysis class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        """Initializes a new instance of the Index class before each test method is run."""
        print("\nTesting Linear Discriminant Analysis", end="", flush=True)

    def setUp(self):  # NOQA D201
        self.cov_class_1 = (
            np.array([[0.0, -1.0], [2.5, 0.7]]) * 2.0
        )  # Covariance matrix for class 1, scaled by 2.0
        self.cov_class_2 = (
            self.cov_class_1.T
        )  # Covariance matrix for class 2, same as class 1 but transposed

        # Generate data
        self.X, self.y = make_sample_data(
            n_samples=1000,
            n_features=2,
            cov_class_1=self.cov_class_1,
            cov_class_2=self.cov_class_2,
            shift=[4, 1],
            seed=1,
        )

    def test_lda(self):
        """Tests the fit and predict methods of the Linear Discriminant Analysis class."""
        lda = LinearDiscriminantAnalysis()
        lda.fit(self.X, self.y)
        y_pred = lda.predict(self.X)
        acc = accuracy_score(self.y, y_pred)
        self.assertGreaterEqual(acc, 0.50)

    def test_lda_svd(self):
        """Tests the fit and predict methods with svd solver of the Linear Discriminant Analysis class."""
        lda = LinearDiscriminantAnalysis(solver="svd")
        lda.fit(self.X, self.y)
        y_pred = lda.predict(self.X)
        acc = accuracy_score(self.y, y_pred)
        self.assertGreater(acc, 0)

    def test_lda_lsqr(self):
        """Tests the fit and predict methods with lsqr solver of the Linear Discriminant Analysis class."""
        lda = LinearDiscriminantAnalysis(solver="lsqr")
        lda.fit(self.X, self.y)
        y_pred = lda.predict(self.X)
        acc = accuracy_score(self.y, y_pred)
        self.assertGreater(acc, 0)

    def test_lda_eigen(self):
        """Tests the fit and predict methods with eigen solver of the Linear Discriminant Analysis class."""
        lda = LinearDiscriminantAnalysis(solver="eigen")
        lda.fit(self.X, self.y)
        y_pred = lda.predict(self.X)
        acc = accuracy_score(self.y, y_pred)
        self.assertGreater(acc, 0)

    def test_lda_bad_solver(self):
        """Tests the fit method with invalid solver of the Linear Discriminant Analysis class."""
        lda = LinearDiscriminantAnalysis(solver="bad_solver")
        with self.assertRaises(ValueError):
            lda.fit(self.X, self.y)

    def test_lda_no_solver(self):
        """Tests the fit method with no solver of the Linear Discriminant Analysis class."""
        lda = LinearDiscriminantAnalysis(solver=None)
        with self.assertRaises(ValueError):
            lda.fit(self.X, self.y)

    def test_lda_no_data(self):
        """Tests the fit method with invalid input."""
        lda = LinearDiscriminantAnalysis()
        with self.assertRaises(ValueError):
            lda.fit(None, None)


class TestQuadraticDiscriminantAnalysis(BaseTest):
    """Unit test for the Quadratic Discriminant Analysis class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        """Initializes a new instance of the Index class before each test method is run."""
        print("\nTesting Quadratic Discriminant Analysis", end="", flush=True)

    def setUp(self):  # NOQA D201
        self.cov_class_1 = (
            np.array([[0.0, -1.0], [2.5, 0.7]]) * 2.0
        )  # Covariance matrix for class 1, scaled by 2.0
        self.cov_class_2 = (
            self.cov_class_1.T
        )  # Covariance matrix for class 2, same as class 1 but transposed

        # Generate data
        self.X, self.y = make_sample_data(
            n_samples=1000,
            n_features=2,
            cov_class_1=self.cov_class_1,
            cov_class_2=self.cov_class_2,
            shift=[4, 1],
            seed=1,
        )

    def test_qda(self):
        """Tests the fit and predict methods of the Quadratic Discriminant Analysis class."""
        qda = QuadraticDiscriminantAnalysis()
        qda.fit(self.X, self.y)
        y_pred = qda.predict(self.X)
        acc = accuracy_score(self.y, y_pred)
        self.assertGreaterEqual(acc, 0.50)

    def test_qda_prior(self):
        """Tests the fit and predict methods with priors parameter of the Quadratic Discriminant Analysis class."""
        qda = QuadraticDiscriminantAnalysis(priors=[0.5, 0.5])
        self.y = self.y.astype(int)  # Ensure class labels are integers
        qda.fit(self.X, self.y)
        y_pred = qda.predict(self.X)
        acc = accuracy_score(self.y, y_pred)
        self.assertGreaterEqual(acc, 0.50)

    def test_qda_reg_param(self):
        """Tests the fit and predict methods with reg_param parameter of the Quadratic Discriminant Analysis class."""
        qda = QuadraticDiscriminantAnalysis(reg_param=0.1)
        qda.fit(self.X, self.y)
        y_pred = qda.predict(self.X)
        acc = accuracy_score(self.y, y_pred)
        self.assertGreaterEqual(acc, 0.50)

    def test_qda_bad_reg_param(self):
        """Tests the fit method with invalid reg_param parameter of the Quadratic Discriminant Analysis class."""
        with self.assertRaises(AssertionError):
            qda = QuadraticDiscriminantAnalysis(reg_param=-0.1)
            qda.fit(self.X, self.y)

    def test_qda_no_data(self):
        """Tests the fit method with invalid input."""
        qda = QuadraticDiscriminantAnalysis()
        with self.assertRaises(ValueError):
            qda.fit(None, None)

    def test_qda_no_priors(self):
        """Tests the fit method with no priors parameter of the Quadratic Discriminant Analysis class."""
        qda = QuadraticDiscriminantAnalysis(priors=None)
        qda.fit(self.X, self.y)
        self.assertEqual(qda.priors, None)

    def test_qda_bad_priors(self):
        """Tests the fit method with invalid priors parameter of the Quadratic Discriminant Analysis class."""
        with self.assertRaises(TypeError):
            qda = QuadraticDiscriminantAnalysis(priors=[0.5, 0.5, 0.5])
            qda.fit(self.X, self.y)


class TestLogisticRegression(BaseTest):
    """Unit test for the Logistic Regression class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        """Initializes a new instance of the Index class before each test method is run."""
        print("\nTesting Logistic Regression", end="", flush=True)

    def setUp(self):  # NOQA D201
        self.cov_class_1 = (
            np.array([[0.0, -1.0], [2.5, 0.7]]) * 2.0
        )  # Covariance matrix for class 1, scaled by 2.0
        self.cov_class_2 = (
            self.cov_class_1.T
        )  # Covariance matrix for class 2, same as class 1 but transposed

        # Generate binary data
        self.X_binary, self.y_binary = make_classification(
            n_samples=300,
            n_features=2,
            n_redundant=0,
            n_informative=2,
            n_classes=2,
            random_state=42,
            n_clusters_per_class=1,
            class_sep=2,
        )

        # Generate multiclass data
        self.X_multiclass, self.y_multiclass = make_classification(
            n_samples=300,
            n_features=2,
            n_redundant=0,
            n_informative=2,
            n_classes=4,
            random_state=42,
            n_clusters_per_class=1,
            class_sep=2,
        )

    def test_fit_predict_binary(self):
        """Tests the fit and predict methods of the Logistic Regression class for binary classification."""
        model = LogisticRegression(learning_rate=0.01, max_iter=1000)
        model.fit(self.X_binary, self.y_binary)
        y_pred = model.predict(self.X_binary)
        acc = accuracy_score(self.y_binary, y_pred)
        self.assertGreaterEqual(acc, 0.50)

    def test_fit_predict_multiclass(self):
        """Tests the fit and predict methods of the Logistic Regression class for multiclass classification."""
        model = LogisticRegression(learning_rate=0.01, max_iter=1000)
        model.fit(self.X_multiclass, self.y_multiclass)
        y_pred = model.predict(self.X_multiclass)
        acc = accuracy_score(self.y_multiclass, y_pred)
        self.assertGreaterEqual(acc, 0.50)

    def test_invalid_fit(self):
        """Tests the fit method with invalid input."""
        model = LogisticRegression()
        with self.assertRaises(ValueError):
            model.fit(None, None)

    def test_invalid_predict(self):
        """Tests the predict method with invalid input."""
        model = LogisticRegression()
        with self.assertRaises(TypeError):
            model.predict(None)


class TestPerceptron(BaseTest):
    """Unit test for the Perceptron class."""

    @classmethod
    def setUpClass(cls):  # NOQA D201
        """Initializes a new instance of the Index class before each test method is run."""
        print("\nTesting Perceptron", end="", flush=True)

    def setUp(self):  # NOQA D201
        self.cov_class_1 = (
            np.array([[0.0, -1.0], [2.5, 0.7]]) * 2.0
        )  # Covariance matrix for class 1, scaled by 2.0
        self.cov_class_2 = (
            self.cov_class_1.T
        )  # Covariance matrix for class 2, same as class 1 but transposed

        # Generate binary data
        self.X_binary, self.y_binary = make_classification(
            n_samples=30,
            n_features=2,
            n_redundant=0,
            n_informative=2,
            n_classes=2,
            random_state=42,
            n_clusters_per_class=1,
            class_sep=2,
        )

        # Generate multiclass data
        self.X_multiclass, self.y_multiclass = make_classification(
            n_samples=30,
            n_features=2,
            n_redundant=0,
            n_informative=2,
            n_classes=4,
            random_state=42,
            n_clusters_per_class=1,
            class_sep=2,
        )

    def test_fit_predict_binary(self):
        """Tests the fit and predict methods of the Perceptron class for binary classification."""
        model = Perceptron(max_iter=1000, learning_rate=0.01)
        model.fit(self.X_binary, self.y_binary)
        y_pred = model.predict(self.X_binary)
        acc = accuracy_score(self.y_binary, y_pred)
        self.assertGreaterEqual(acc, 0.50)

    def test_fit_predict_multiclass(self):
        """Tests the fit and predict methods of the Perceptron class for multiclass classification."""
        model = Perceptron(max_iter=1000, learning_rate=0.01)
        model.fit(self.X_multiclass, self.y_multiclass)
        y_pred = model.predict(self.X_multiclass)
        acc = accuracy_score(self.y_multiclass, y_pred)
        self.assertGreaterEqual(acc, 0.50)

    def test_invalid_fit(self):
        """Tests the fit method with invalid input."""
        model = Perceptron()
        with self.assertRaises(ValueError):
            model.fit(None, None)

    def test_invalid_predict(self):
        """Tests the predict method with invalid input."""
        model = Perceptron()
        with self.assertRaises(TypeError):
            model.predict(None)


if __name__ == "__main__":
    unittest.main()



import os
import sys
import unittest

# Change the working directory to the parent directory to allow importing the package.
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from sega_learn.utils import *
from tests.utils import BaseTest


class TestImportsUtils(BaseTest):
    """Tests that the Utils subpackage can be imported correctly.

    Methods:
    - setUpClass: Initializes a new instance of the Index class before each test method is run.
    - test_individual_imports: Tests that each module in the segadb package can be imported individually.
    - test_wildcard_import: Tests that the segadb package can be imported using a wildcard import.
    """

    @classmethod
    def setUpClass(cls):  # NOQA D201
        print("\nTesting Imports - Utils", end="", flush=True)

    def test_individual_imports(self):
        """Tests that each module in the segadb package can be imported individually."""
        from sega_learn.utils import PCA as pca
        from sega_learn.utils import SMOTE as smote
        from sega_learn.utils import SVD as svd
        from sega_learn.utils import AnimationBase as anb
        from sega_learn.utils import Augmenter as augmenter
        from sega_learn.utils import ClassificationAnimation as canb
        from sega_learn.utils import DataPrep as dp
        from sega_learn.utils import ForcastingAnimation as fanb
        from sega_learn.utils import ForecastRegressor as fr
        from sega_learn.utils import GridSearchCV as gscv
        from sega_learn.utils import Metrics as metrics
        from sega_learn.utils import ModelSelectionUtility as msu
        from sega_learn.utils import PolynomialTransform as plt
        from sega_learn.utils import RandomOverSampler as ros
        from sega_learn.utils import RandomSearchCV as rscv
        from sega_learn.utils import RandomUnderSampler as rus
        from sega_learn.utils import RegressionAnimation as ranb
        from sega_learn.utils import Scaler as scaler
        from sega_learn.utils import VotingClassifier as vc
        from sega_learn.utils import VotingRegressor as vr
        from sega_learn.utils import check_is_fitted as cif
        from sega_learn.utils import check_X_y as cxy
        from sega_learn.utils import make_blobs as mkblobs
        from sega_learn.utils import make_classification as mkcls
        from sega_learn.utils import make_regression as mkreg
        from sega_learn.utils import make_time_series as mkts
        from sega_learn.utils import normalize as norm
        from sega_learn.utils import one_hot_encode as ohe
        from sega_learn.utils import train_test_split as tts

        assert plt is not None
        assert dp is not None
        assert vr is not None
        assert vc is not None
        assert msu is not None
        assert gscv is not None
        assert rscv is not None
        assert metrics is not None
        assert ros is not None
        assert rus is not None
        assert smote is not None
        assert augmenter is not None
        assert pca is not None
        assert svd is not None
        assert mkreg is not None
        assert mkcls is not None
        assert mkblobs is not None
        assert mkts is not None
        assert tts is not None
        assert ohe is not None
        assert norm is not None
        assert scaler is not None
        assert fr is not None
        assert anb is not None
        assert ranb is not None
        assert fanb is not None
        assert canb is not None
        assert cif is not None
        assert cxy is not None

    def test_wildcard_import(self):
        """Tests that the segadb package can be imported using a wildcard import."""
        assert PolynomialTransform is not None
        assert DataPrep is not None
        assert VotingRegressor is not None
        assert ModelSelectionUtility is not None
        assert GridSearchCV is not None
        assert RandomSearchCV is not None
        assert Metrics is not None
        assert RandomOverSampler is not None
        assert RandomUnderSampler is not None
        assert SMOTE is not None
        assert Augmenter is not None
        assert PCA is not None
        assert SVD is not None
        assert make_regression is not None
        assert make_classification is not None
        assert make_blobs is not None
        assert make_time_series is not None
        assert train_test_split is not None
        assert one_hot_encode is not None
        assert normalize is not None
        assert Scaler is not None
        assert AnimationBase is not None
        assert RegressionAnimation is not None
        assert ForcastingAnimation is not None
        assert ClassificationAnimation is not None

    def test_AnimationBase(self):  # NOQA D201
        from sega_learn.utils import AnimationBase

        assert AnimationBase is not None

    def test_RegressionAnimation(self):  # NOQA D201
        from sega_learn.utils import RegressionAnimation

        assert RegressionAnimation is not None

    def test_ForcastingAnimation(self):  # NOQA D201
        from sega_learn.utils import ForcastingAnimation

        assert ForcastingAnimation is not None

    def test_ClassificationAnimation(self):  # NOQA D201
        from sega_learn.utils import ClassificationAnimation

        assert ClassificationAnimation is not None

    def test_PCA(self):  # NOQA D201
        from sega_learn.utils import PCA

        assert PCA is not None

    def test_SMOTE(self):  # NOQA D201
        from sega_learn.utils import SMOTE

        assert SMOTE is not None

    def test_SVD(self):  # NOQA D201
        from sega_learn.utils import SVD

        assert SVD is not None

    def test_Augmenter(self):  # NOQA D201
        from sega_learn.utils import Augmenter

        assert Augmenter is not None

    def test_DataPrep(self):  # NOQA D201
        from sega_learn.utils import DataPrep

        assert DataPrep is not None

    def test_GridSearchCV(self):  # NOQA D201
        from sega_learn.utils import GridSearchCV

        assert GridSearchCV is not None

    def test_Metrics(self):  # NOQA D201
        from sega_learn.utils import Metrics

        assert Metrics is not None

    def test_ModelSelectionUtility(self):  # NOQA D201
        from sega_learn.utils import ModelSelectionUtility

        assert ModelSelectionUtility is not None

    def test_PolynomialTransform(self):  # NOQA D201
        from sega_learn.utils import PolynomialTransform

        assert PolynomialTransform is not None

    def test_RandomOverSampler(self):  # NOQA D201
        from sega_learn.utils import RandomOverSampler

        assert RandomOverSampler is not None

    def test_RandomSearchCV(self):  # NOQA D201
        from sega_learn.utils import RandomSearchCV

        assert RandomSearchCV is not None

    def test_RandomUnderSampler(self):  # NOQA D201
        from sega_learn.utils import RandomUnderSampler

        assert RandomUnderSampler is not None

    def test_Scaler(self):  # NOQA D201
        from sega_learn.utils import Scaler

        assert Scaler is not None

    def test_VotingRegressor(self):  # NOQA D201
        from sega_learn.utils import VotingRegressor

        assert VotingRegressor is not None

    def test_ForecastRegressor(self):  # NOQA D201
        from sega_learn.utils import ForecastRegressor

        assert ForecastRegressor is not None

    def test_make_blobs(self):  # NOQA D201
        from sega_learn.utils import make_blobs

        assert make_blobs is not None

    def test_make_classification(self):  # NOQA D201
        from sega_learn.utils import make_classification

        assert make_classification is not None

    def test_make_regression(self):  # NOQA D201
        from sega_learn.utils import make_regression

        assert make_regression is not None

    def test_make_time_series(self):  # NOQA D201
        from sega_learn.utils import make_time_series

        assert make_time_series is not None

    def test_normalize(self):  # NOQA D201
        from sega_learn.utils import normalize

        assert normalize is not None

    def test_one_hot_encode(self):  # NOQA D201
        from sega_learn.utils import one_hot_encode

        assert one_hot_encode is not None

    def test_train_test_split(self):  # NOQA D201
        from sega_learn.utils import train_test_split

        assert train_test_split is not None

    def test_check_is_fitted(self):  # NOQA D201
        from sega_learn.utils import check_is_fitted

        assert check_is_fitted is not None

    def test_check_X_y(self):  # NOQA D201
        from sega_learn.utils import check_X_y

        assert check_X_y is not None


if __name__ == "__main__":
    unittest.main()



from .neuralNetworkBase import NeuralNetworkBase
from .schedulers import lr_scheduler_exp, lr_scheduler_plateau, lr_scheduler_step

__all__ = [
    "lr_scheduler_exp",
    "lr_scheduler_plateau",
    "lr_scheduler_step",
    "NeuralNetworkBase",
]

try:
    from .layers_jit_unified import JITLayer
    from .loss_jit import JITBCEWithLogitsLoss, JITCrossEntropyLoss
    from .neuralNetworkNumbaBackend import NumbaBackendNeuralNetwork
    from .numba_utils import *
    from .optimizers_jit import JITAdadeltaOptimizer, JITAdamOptimizer, JITSGDOptimizer

    __all__.extend(
        [
            "JITAdamOptimizer",
            "JITSGDOptimizer",
            "JITAdadeltaOptimizer",
            "JITBCEWithLogitsLoss",
            "JITCrossEntropyLoss",
            "NumbaBackendNeuralNetwork",
            "JITLayer",
        ]
    )
except:
    pass



import numpy as np


class lr_scheduler_step:
    """
    Learning rate scheduler class for training neural networks.
    Reduces the learning rate by a factor of lr_decay every lr_decay_epoch epochs.
    Args:
        optimizer (Optimizer): The optimizer to adjust the learning rate for.
        lr_decay (float, optional): The factor to reduce the learning rate by. Defaults to 0.1.
        lr_decay_epoch (int, optional): The number of epochs to wait before decaying the learning rate. Defaults to 10
    """

    def __init__(self, optimizer, lr_decay=0.1, lr_decay_epoch=10):
        self.optimizer = optimizer
        self.lr_decay = lr_decay
        self.lr_decay_epoch = lr_decay_epoch

    def __repr__(self):
        return f"StepLR({self.optimizer}, lr_decay={self.lr_decay}, lr_decay_epoch={self.lr_decay_epoch})"

    def step(self, epoch):
        """
        Adjusts the learning rate based on the current epoch. Decays the learning rate by lr_decay every lr_decay_epoch epochs.
        Args:
            epoch (int): The current epoch number.
        Returns: None
        """
        if epoch % self.lr_decay_epoch == 0 and epoch > 0:
            self.optimizer.learning_rate *= self.lr_decay
            return f"  --Decaying learning rate stepped at epoch {epoch} from {self.optimizer.learning_rate} to {self.optimizer.learning_rate * self.lr_decay}"

    def reduce(self):
        self.optimizer.learning_rate *= self.lr_decay
        return f"  --Decaying learning rate from {self.optimizer.learning_rate} to {self.optimizer.learning_rate * self.lr_decay}"


class lr_scheduler_exp:
    """
    Learning rate scheduler class for training neural networks.
    Reduces the learning rate exponentially by lr_decay every lr_decay_epoch epochs.
    """

    def __init__(self, optimizer, lr_decay=0.1, lr_decay_epoch=10):
        self.optimizer = optimizer
        self.lr_decay = lr_decay
        self.lr_decay_epoch = lr_decay_epoch

    def __repr__(self):
        return f"ExponentialLR({self.optimizer}, lr_decay={self.lr_decay}, lr_decay_epoch={self.lr_decay_epoch})"

    def step(self, epoch):
        """
        Adjusts the learning rate based on the current epoch. Decays the learning rate by lr_decay every lr_decay_epoch epochs.
        Args:
            epoch (int): The current epoch number.
        Returns: None
        """
        if epoch % self.lr_decay_epoch == 0 and epoch > 0:
            self.optimizer.learning_rate *= np.exp(-self.lr_decay * epoch)
            return f"  --Decaying learning rate exponentially at epoch {epoch} from {self.optimizer.learning_rate} to {self.optimizer.learning_rate * np.exp(-self.lr_decay * epoch)}"

    def reduce(self):
        self.optimizer.learning_rate *= np.exp(-self.lr_decay)
        return f"  --Decaying learning rate exponentially from {self.optimizer.learning_rate} to {self.optimizer.learning_rate * np.exp(-self.lr_decay)}"


class lr_scheduler_plateau:
    """
    A custom learning rate scheduler that adjusts the learning rate based on the plateau of the loss function.
    Args:
        lr_scheduler (object): The learning rate scheduler object.
        patience (int): The number of epochs to wait for improvement before reducing the learning rate. Default is 5.
        threshold (float): The minimum improvement threshold required to update the best loss. Default is 0.01.
    Methods:
        step(loss): Updates the learning rate based on the loss value.
    """

    def __init__(self, lr_scheduler, patience=5, threshold=0.01):
        self.lr_scheduler = lr_scheduler
        self.patience = patience
        self.threshold = threshold
        self.best_loss = float("inf")
        self.wait = 0

    def __repr__(self):
        return f"PlateauLR({self.lr_scheduler}, patience={self.patience}, threshold={self.threshold})"

    def step(self, epoch, loss):
        """
        Updates the learning rate based on the loss value.
        Args:
            loss (float): The current loss value.
        """
        if loss < self.best_loss - self.threshold:
            self.best_loss = loss
            self.wait = 0
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.lr_scheduler.reduce()
                self.wait = 0
                return "  --Plateau learning rate scheduler triggered, reducing learning rate"



from .arima import ARIMA, SARIMA, SARIMAX
from .decomposition import AdditiveDecomposition, MultiplicativeDecomposition
from .exponential_smoothing import (
    DoubleExponentialSmoothing,
    SimpleExponentialSmoothing,
    TripleExponentialSmoothing,
)
from .moving_average import (
    ExponentialMovingAverage,
    SimpleMovingAverage,
    WeightedMovingAverage,
)

__all__ = [
    "ARIMA",
    "SARIMA",
    "SARIMAX",
    "SimpleMovingAverage",
    "WeightedMovingAverage",
    "ExponentialMovingAverage",
    "SimpleExponentialSmoothing",
    "DoubleExponentialSmoothing",
    "TripleExponentialSmoothing",
    "AdditiveDecomposition",
    "MultiplicativeDecomposition",
]



import numpy as np
from numba import njit, prange


@njit(fastmath=True)
def _linearSVC_minibatches(X, y, w, b, C, beta, learning_rate, batch_size):
    """Process all mini-batches for LinearSVC using gradient descent with momentum.

    Args:
        X (ndarray): Training data of shape (n_samples, n_features).
        y (ndarray): Target labels of shape (n_samples,).
        w (ndarray): Weight vector of shape (n_features,).
        b (float): Bias term.
        C (float): Regularization parameter.
        beta (float): Momentum factor.
        learning_rate (float): Learning rate for gradient descent.
        batch_size (int): Size of each mini-batch.

    Returns:
        w (ndarray): Updated weight vector.
        b (float): Updated bias term.
        dw (ndarray): Gradient of the weight vector.
        db (float): Gradient of the bias term.
    """
    n_samples, n_features = X.shape
    momentum_w = np.zeros(n_features)
    momentum_b = 0.0

    # Calculate the number of batches
    num_batches = (n_samples + batch_size - 1) // batch_size  # Ceiling division

    for batch_idx in prange(num_batches):
        start = batch_idx * batch_size
        end = min(start + batch_size, n_samples)
        X_batch = X[start:end]
        y_batch = y[start:end]

        # Compute the margin
        margin = y_batch * (np.dot(X_batch, w) + b)

        # Compute the hinge loss gradient
        violated_indices = margin < 1
        dw = C * w  # L2 regularization gradient
        db = 0.0

        if np.any(violated_indices):
            X_violated = X_batch[violated_indices]
            y_violated = y_batch[violated_indices]
            dw -= np.sum(X_violated * y_violated[:, np.newaxis], axis=0) / batch_size
            db -= np.sum(y_violated) / batch_size

        # Apply momentum
        momentum_w = beta * momentum_w + (1 - beta) * dw
        momentum_b = beta * momentum_b + (1 - beta) * db

        # Update weights and bias
        w -= learning_rate * momentum_w
        b -= learning_rate * momentum_b

    return w, b, dw, db


@njit(fastmath=True)
def _linearSVR_minibatches(X, y, w, b, C, beta, learning_rate, batch_size, epsilon):
    """Process all mini-batches for LinearSVR using gradient descent with momentum.

    Args:
        X (ndarray): Training data of shape (n_samples, n_features).
        y (ndarray): Target values of shape (n_samples,).
        w (ndarray): Weight vector of shape (n_features,).
        b (float): Bias term.
        C (float): Regularization parameter.
        beta (float): Momentum factor.
        learning_rate (float): Learning rate for gradient descent.
        batch_size (int): Size of each mini-batch.
        epsilon (float): Epsilon-insensitive loss parameter.

    Returns:
        w (ndarray): Updated weight vector.
        b (float): Updated bias term.
        dw (ndarray): Gradient of the weight vector.
        db (float): Gradient of the bias term.
    """
    n_samples, n_features = X.shape
    momentum_w = np.zeros(n_features)
    momentum_b = 0.0

    # Calculate the number of batches
    num_batches = (n_samples + batch_size - 1) // batch_size  # Ceiling division

    for batch_idx in prange(num_batches):
        start = batch_idx * batch_size
        end = start + batch_size
        X_batch = X[start:end]
        y_batch = y[start:end]

        # Compute predictions
        predictions = np.dot(X_batch, w) + b

        # Compute errors
        errors = y_batch - predictions

        # Initialize gradients
        dw = C * w  # L2 regularization gradient
        db = 0.0

        # Positive epsilon-insensitive loss
        pos_idx = errors > epsilon
        if np.any(pos_idx):
            dw -= np.sum(X_batch[pos_idx], axis=0) / batch_size
            db -= np.sum(np.ones(np.sum(pos_idx))) / batch_size

        # Negative epsilon-insensitive loss
        neg_idx = errors < -epsilon
        if np.any(neg_idx):
            dw += np.sum(X_batch[neg_idx], axis=0) / batch_size
            db += np.sum(np.ones(np.sum(neg_idx))) / batch_size

        # Apply momentum
        momentum_w = beta * momentum_w + (1 - beta) * dw
        momentum_b = beta * momentum_b + (1 - beta) * db

        # Update weights and bias
        w -= learning_rate * momentum_w
        b -= learning_rate * momentum_b

    return w, b, dw, db
